
__author__ = "Andrea Tramacere"

import math as m


from numpy import polyfit,polyval


import numpy as np

from .frame_converter import convert_nu_to_blob

from . import jetkernel_models_dic as Model_dic

import os

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

if on_rtd == True:
    try:
        from .jetkernel import jetkernel as BlazarSED
    except ImportError:
        from .mock import jetkernel as BlazarSED
else:
    from .jetkernel import jetkernel as BlazarSED


from .jet_model import Jet

from .sed_shaper import index_array,peak_values

from .output import section_separator,WorkPlace,makedir

from .utils import *

__all__=['ObsConstrain','check_boundaries', 'check_t_var',
         'constr_B_from_nu_peaks','constr_R_from_CD','find_B_from_nu_p_S','find_gamma0',
         'find_gamma_3p_SSC','find_gamma_Synch','find_HE_cut_off','find_s','find_s1',
         'find_turn_over','get_Comp_factor','get_R_tvar','get_U_Sync_from_Ph',
         'rescale_Ne','set_gmin_from_nu_cut_IR']



class ObsConstrain(object):
    """
    doc
    """
    def __init__(self,B_range=None,
                 distr_e=None,
                 t_var_sec=None,
                 nu_cut_IR=None,
                 beaming=None,
                 theta=None,
                 bulk_factor=None,
                 obj_class=None,
                 #restframe=None,
                 z=None,
                 obspar_workplace=None,
                 **keywords):

        if 'SEDShape' not in keywords :

            self.indices=index_array()
            
            keys = sorted(keywords.keys())
            for kw in keys:
                print("kw",kw)
                if kw =='indices':
                    self.indices=keywords[kw]
                    
            for index in self.indices.idx_array:
                index.show_val()
            
            self.beta_S=keywords['beta_S']
            self.nu_p_S_obs=keywords['nu_p_S_obs']
            self.nu_p_IC_obs=keywords['nu_p_IC_obs']
            self.nuFnu_p_S_obs=keywords['nuFnu_p_S_obs']
            self.nuFnu_p_IC_obs=keywords['nuFnu_p_IC_obs']
            self.S_nu_max=keywords['S_nu_max']
            
            self.z=z
            self.class_obj=obj_class

        else:
            
            self.SEDShape=keywords['SEDShape']
            self.indices=self.SEDShape.indices
           
            self.beta_S = -self.SEDShape.S_peak.curvature
            self.nu_p_S_obs = np.power(10., self.SEDShape.S_peak.nu_p_val)
            self.nuFnu_p_S_obs = np.power(10., self.SEDShape.S_peak.nuFnu_p_val)
            self.S_nu_max=np.power(10., self.SEDShape.S_nu_max)
            self.S_LE_slope=self.SEDShape.S_LE_slope
            self.nu_p_IC_obs  = np.power(10., self.SEDShape.IC_peak.nu_p_val)
            self.nuFnu_p_IC_obs = np.power(10., self.SEDShape.IC_peak.nuFnu_p_val)

            self.class_obj = self.SEDShape.obj_class
            self.z=self.SEDShape.sed_data.z
            self.rest_frame='obs'
            check_frame(self.rest_frame)

        if self.z<0.:
            raise  RuntimeError('redshift value must be>0, please check z in sed_data or z in class constructor parameter')
        
        
            
        self.beaming=beaming
        self.theta=theta
        self.bulk_factor=bulk_factor

        if self.theta is None and self.bulk_factor is None and self.beaming is not None:
            self.beaming_expr='delta'
        elif self.theta is not None and self.bulk_factor is not None and self.beaming is None:
            self.beaming_expr='bulk_theta'
        else:
            raise RuntimeError('''either you provide the beaming value, or both theta and the bulk factor values ''')
                
        
        self.B_min=B_range[0]
        self.B_max=B_range[1]
        self.B_start = (self.B_min + self.B_max)/2.0

        self.t_var_sec=t_var_sec
        
        self.distr_e=distr_e
        
        self.nu_cut_IR=nu_cut_IR

        if obspar_workplace is None:
            obspar_workplace=WorkPlace()
            self.out_dir=obspar_workplace.out_dir
            self.flag=obspar_workplace.flag
        else:
            self.out_dir=obspar_workplace.out_dir
            self.flag=obspar_workplace.flag

    def constrain_SSC_model(self,name=None,jet_model=None,params_grid_size=10,electron_distribution_log_values=False,silent=False):

        """
        constarin SSC model paramters
        """
        print(section_separator)
        print("***  constrains parameters from observable ***")
        print()

        model=self.get_model_constraint(name=name,
                                        jet_model=jet_model,
                                        params_grid_size=params_grid_size,
                                        electron_distribution_log_values=electron_distribution_log_values,
                                        silent=silent)
        print()
        print(section_separator)

        return model
    
    
    
    
    def constrain_SSC_EC_model(self,name=None,
                               jet_model=None,
                               EC_componets_list=['EC_BLR'],
                               params_grid_size=10,
                               electron_distribution_log_values=False,
                               R_H=None,
                               silent=False,
                               R_H_within_BLR=False,
                               R_H_within_DT=False,
                               disk_type='BB'):
        """
        constarin SSC model paramters
        """
        
         
        
        print(section_separator)
        print ("***  constrains parameters from observable ***")
        print()
        
        
        model=self.get_model_constraint(name=name,
                                        jet_model=jet_model,
                                        EC_componets_list=EC_componets_list,
                                        params_grid_size=params_grid_size,
                                        electron_distribution_log_values=electron_distribution_log_values,
                                        silent=silent,
                                        R_H=R_H,
                                        R_H_within_BLR=R_H_within_BLR,
                                        R_H_within_DT=R_H_within_DT,
                                        disk_type=disk_type)
        
         

        print()
        print (section_separator)

        return model
     
     
     
    
    def get_model_constraint(self,
                             name=None,
                             jet_model=None,
                             EC_componets_list=None,
                             params_grid_size=10,
                             electron_distribution_log_values=False,
                             silent=False,
                             R_H=None,
                             R_H_within_BLR=False,
                             R_H_within_DT=False,
                             disk_type='BB'):
        


        #out_dir='%s/obs_constrain_%s/'%(self.out_dir,name)
        #makedir(out_dir)

        if silent is False:
            print(section_separator)
            print("---> ***  emitting region parameters  ***")

        if jet_model is None:
            if self.beaming_expr=='delta':
                jet_model=Jet(name=name, electron_distribution=self.distr_e,electron_distribution_log_values=electron_distribution_log_values)
            elif self.beaming_expr=='bulk_theta':
                jet_model=Jet(name=name, electron_distribution=self.distr_e,beaming_expr='bulk_theta',electron_distribution_log_values=electron_distribution_log_values)

            else:
                raise RuntimeError('''wrong beaming_expr value=%s, allowed 'delta' or 'bulk_theta' '''%self.beaming_expr)
        if jet_model.emitters_distribution.spectral_type not in jet_model.emitters_distribution.spectral_types_obs_constrain():
            raise RuntimeError('''to use osb constrain the spectral type of emitters has to be ''' %jet_model.emitters_distribution.spectral_types_obs_constrain())

        if R_H is not None:
            jet_model.set_par('R_H',val=R_H)
        self.emitters_distr_spectral_type=jet_model.emitters_distribution.spectral_type

        nu_p_EC_seed_field=None
        if EC_componets_list is not None:
            jet_model.add_EC_component(EC_componets_list,disk_type=disk_type)
            if hasattr(jet_model.parameters, 'L_Disk'):
                jet_model.set_par('L_Disk',val=self.SEDShape.L_Disk)
            if hasattr(jet_model.parameters, 'T_Disk'):
                jet_model.set_par('T_Disk', val=self.SEDShape.T_Disk)
                if silent is False:
                    print('---> EC set L_D ',jet_model.parameters.L_Disk.val)
                    print('---> EC set T_D', jet_model.parameters.T_Disk.val)
                if hasattr(jet_model.parameters, 'R_BLR_in'):
                    R_BLR_in=1E17*np.sqrt(self.SEDShape.L_Disk/1E45)
                    jet_model.set_par('R_BLR_in', val=R_BLR_in)
                    jet_model.set_par('R_BLR_out', val=R_BLR_in*2)

                    if silent is False:
                        print('---> EC set R_BLR_in', jet_model.parameters.R_BLR_in.val)
                        print('---> EC set R_BLR_out', jet_model.parameters.R_BLR_out.val)
                    if jet_model.parameters.R_H.val > R_BLR_in and R_H_within_BLR is True:
                        jet_model.parameters.R_H.val = R_BLR_in * 0.8
                        if silent is False:
                            print('---> moved R_H within BLR to R_H=%e'%jet_model.parameters.R_H.val)
                if hasattr(jet_model.parameters, 'R_DT'):
                    R_DT=2.5E18*np.sqrt(self.SEDShape.L_Disk/1E45)
                    jet_model.set_par('R_DT', val = R_DT)
                    if silent is False:
                        print('---> EC set R_DT', jet_model.parameters.R_DT.val)
                    if jet_model.parameters.R_H.val > R_DT and R_H_within_DT is True:
                        jet_model.parameters.R_H.val = R_DT * 0.8
                        if silent is False:
                            print('---> moved R_H within DT to R_H=%e'%jet_model.parameters.R_H.val)


            nu_p_EC_seed_field=self.SEDShape.nu_p_Disk

        if silent is False:
            print()
        #setting the Jet object 
        #path_initial=jet_model.get_path()
        #flag_initial=jet_model.get_flag()
        
        #jet_model.set_path(out_dir)
        #jet_model.set_flag(name)

        IC_nu_size_initial=jet_model.get_IC_nu_size()
        nu_seed_size_initial=jet_model.get_seed_nu_size()


        jet_model.set_IC_nu_size(50)
        jet_model.set_seed_nu_size(50)

        #jet_model.show_pars()
        
        #beaming

       
       
        #SEtting emetting_region parameters
        
        #!!Controlla beaming max
        #beaming

        if self.beaming_expr=='delta':
            beaming_par=jet_model.get_par_by_type('beaming')
            beaming_par.set(val=self.beaming)
            #if silent is False:
            #    print("--->",beaming_par.get_description())
            
           
        elif self.beaming_expr=='bulk_theta': 
            bulk_factor_par=jet_model.get_par_by_type('jet-bulk-factor')
            theta_par=jet_model.get_par_by_type('jet-viewing-angle')
            theta_par.set(val=self.theta)
            bulk_factor_par.set(val=self.bulk_factor)
            #if silent is False:
            #    print("--->",theta_par.get_description())
            #    print("--->",bulk_factor_par.get_description())
            #self.beaming=jet_model.get_beaming(theta=self.theta,bulk_factor=self.bulk_factor,beaming=-1.0)
            self.beaming = jet_model.get_beaming()
            if silent is False:
                print("---> setting beaming  to",self.beaming)
                print()
        else:
            raise RuntimeError('''wrong beaming_expr value=%s, allowed 'delta' or 'bulk_theta' '''%self.beaming_expr)
        
            
        #z
        z_par=jet_model.get_par_by_type('redshift')
        if z_par is not None:
            if silent is False:
                print ("---> setting par type redshift, corresponding to par %s"%(z_par.name))
                print()
            z_par.set(val=self.z)
            #if silent is False:
            #    print("---> ",z_par.get_description())
            #    print()
        #B
        B_par=jet_model.get_par_by_type('magnetic_field')
        B_par.set(val=self.B_start)
        if B_par is not None:
            if silent is False:
                print("---> setting par type magnetic_field, corresponding to par %s=%e"%(B_par.name,B_par.val))
                print()



        #R
        R_tvar,completed=get_R_tvar(self.beaming,self.t_var_sec,+self.z)

        R_par=jet_model.get_par_by_type('region_size')
        R_par.set(val=set_lin_log_val(R_par, R_tvar))
        if R_par is not None and completed is True:
            if silent is False:

                print("---> setting par type region_size, corresponding to par %s=%e"%(R_par.name,R_par.val))
                print("---> completed", completed)
                print()



        if silent is False:
            print()
            print("---> *** electron distribution parameters ***")
            print('---> emitters distribution spectral type', jet_model.emitters_distribution.spectral_type)
            print('---> emitters distribution name', jet_model.emitters_distribution.name)
            print()

        #r
        curvature_par=jet_model.get_par_by_type('spectral_curvature')
        if curvature_par is not None:
        
            if self.beta_S is not None:
                curvature_par.set(val=self.beta_S*5.0)
                if silent is False:
                    print("---> r elec. spec. curvature =%e"%curvature_par.val)
            else:
                curvature_par.set(val=0.4*5.0)
                if silent is False:
                    print("---> beta Sync not provided, using 0.4, corresponding to r=2.0")
            if silent is False:
                print("---> setting par type curvature, corresponding to par %s"%(curvature_par.name))
                #print("---> ",curvature_par.get_description())
                print()
        
        #s
        LE_spectral_slope_par=jet_model.get_par_by_type('LE_spectral_slope')
        if LE_spectral_slope_par is not None:
            (s_Plank,s_X,s_Fermi,index_s),completed=find_s(self.class_obj,self.nu_p_S_obs,self.S_LE_slope,self.indices,silent=silent)
            LE_spectral_slope_par.set(val=index_s)
            if silent is False:
                print("---> setting par type LE_spectral_slope, corresponding to par %s"%(LE_spectral_slope_par.name))
                print("---> task completed",completed)
                #print("---> ",LE_spectral_slope_par.get_description())
                print()


        #s1
        HE_spectral_slope_par=jet_model.get_par_by_type('HE_spectral_slope')
        if HE_spectral_slope_par is not None:
            index_s1,completed=find_s1(self.class_obj,self.indices)
            HE_spectral_slope_par.set(val=index_s1)
            if silent is False:
                print("---> setting par type LE_spectral_slope, corresponding to par %s"%(HE_spectral_slope_par.name))
                print("---> task completed", completed)
                #print("---> ",HE_spectral_slope_par.get_description())
                print()

        gamma_3p_Sync,completed=find_gamma_Synch(self.nu_p_S_obs,self.rest_frame,B_par.val,self.beaming,z_par.val)
        if silent is False:
            print ("---> setting gamma_3p_Sync= %e, assuming B=%e"%(gamma_3p_Sync,B_par.val))
            print ("---> task completed",completed)
            print()
        
        #gmax start
        gmax, completed=find_HE_cut_off(self.emitters_distr_spectral_type,self.S_nu_max,self.rest_frame,B_par.val,self.beaming,z_par.val)
        if silent is False:
            print("---> gamma_max=%e from nu_max_Sync= %e, using B=%e"%(gmax,self.S_nu_max,B_par.val))
            print("---> task completed", completed)
        HE_cut_off_par=jet_model.get_par_by_type('high-energy-cut-off')
        if HE_cut_off_par is not None:
            HE_cut_off_par.set(val=set_lin_log_val(HE_cut_off_par,gmax))
            if silent is False:
                print("---> setting par type high-energy-cut-off, corresponding to par %s=%e"%(HE_cut_off_par.name,HE_cut_off_par.val))
                #print("---> ",HE_cut_off_par.get_description())
                print()
        
        #gmin start
        gmin,completed= set_gmin_from_nu_cut_IR(self.nu_cut_IR,self.rest_frame,B_par.val,self.beaming,z_par.val)
        par_LE_cut_off_par=jet_model.get_par_by_type('low-energy-cut-off')
        if par_LE_cut_off_par is not None:
            par_LE_cut_off_par.set(val=set_lin_log_val(par_LE_cut_off_par,gmin))
            if silent is False:
                print("---> setting par type low-energy-cut-off, corresponding to par %s=%e"%(par_LE_cut_off_par.name,par_LE_cut_off_par.val))
                print("---> task completed", completed)
                #print("---> ",par_LE_cut_off_par.get_description())
                print()

        #turn-over-energy from gamma_3p_Sync
        turn_over_par=jet_model.get_par_by_type('turn-over-energy')
        if turn_over_par is not None:
            _t,completed=find_turn_over(jet_model,self.emitters_distr_spectral_type,gamma_3p_Sync)
            turn_over_par.set(val=set_lin_log_val(turn_over_par,_t))

            if silent is False:
                print("---> setting par type turn-over energy, corresponding to par %s=%e"%(turn_over_par.name,turn_over_par.val))
                print("---> task completed", completed)
                print("---> using gamma_3p_Sync=",gamma_3p_Sync)
                #print("---> ",turn_over_par.get_description())
                print()

        #turn-over-energy from gamma_3p_SSC
        gamma_3p_SSC,completed= find_gamma_3p_SSC(self.nu_p_S_obs,self.nu_p_IC_obs,self.rest_frame,gamma_3p_Sync,self.beaming,z_par.val,nu_p_EC_seed_field=nu_p_EC_seed_field,silent=silent)
        if silent is False:
            print("---> determine gamma_3p_SSCc= %e"%gamma_3p_SSC)
            print("---> task completed", completed)
            print()
       
        if turn_over_par is not None:
            _t,completed=find_turn_over(jet_model,self.emitters_distr_spectral_type,gamma_3p_SSC)
            turn_over_par.set(val=set_lin_log_val(turn_over_par,_t))

            if silent is False:
                print("---> setting par type turn-over energy, corresponding to par %s=%e"%(turn_over_par.name,turn_over_par.val))
                print("---> task completed", completed)
                print("---> using gamma_3p_SSC=%e"%gamma_3p_SSC)
                #print("---> ",turn_over_par.get_description())
                print()

        if silent is False:
            print ()
       
       
       
        #N
        N_par=jet_model.get_par_by_type('emitters_density')
        if N_par is not None:
            completed=rescale_Ne(jet_model,self.nuFnu_p_S_obs,self.nu_p_S_obs,self.rest_frame)
            if silent is False:
                print("---> setting par type emitters_density, corresponding to par %s"%(N_par.name))
                print("---> to N=%e" % (N_par.val))
                print("---> task completed",completed)
            if silent is False:
                #print('--->',N_par.get_description())
                print()

        B,B=find_B_from_nu_p_S(self.nu_p_S_obs,gamma_3p_SSC,self.rest_frame,self.beaming,z_par.val)
        if silent is False:
            print("---> setting B from nu_p_S to B=%e"%B)
            print("---> to B=%e" %B)

        if silent is False:
            print("---> setting B from best matching of nu_p_IC")
            print()

        if B_par is not None:
            if EC_componets_list is None:
                (B_from_nu_peaks, failed), completed = constr_B_from_nu_peaks (jet_model,self.nu_p_S_obs,self.nu_p_IC_obs,self.rest_frame,self.B_min,self.B_max,self.beaming,params_grid_size,silent=silent)
            else:
                (B_from_nu_peaks, failed), completed = constr_B_from_nu_peaks (jet_model,self.nu_p_S_obs,self.nu_p_IC_obs,self.rest_frame,self.B_min,self.B_max,self.beaming,params_grid_size,EC=True,silent=silent)

            if completed is True and failed is False:
                B_par.set(val=B_from_nu_peaks)

            if silent is False:
                print("---> setting par type magnetic_field, corresponding to par %s"%(B_par.name))
                #print("---> ",B_par.get_description())
                print("---> task completed ",completed)

            if failed is False and completed is True:
                if silent is False:
                    print ("---> best B found: %e"%B_par.val)
            else:
                if silent is False:
                    print("---> constrain failed, B set to: %e"%B_par.val)
                    print()

            if silent is False:
                print()
                print("---> update pars for new B ")
            
            #update gmin from new B
            gmin,completed= set_gmin_from_nu_cut_IR(self.nu_cut_IR,self.rest_frame,B_par.val,self.beaming,z_par.val)
            if par_LE_cut_off_par is not None:
                par_LE_cut_off_par.set(val=set_lin_log_val(par_LE_cut_off_par,gmin))

                if silent is False:
                    print("---> setting par type low-energy-cut-off, corresponding to par %s"%(par_LE_cut_off_par.name))
                    print("---> task completed", completed)
                    print("---> set to %e"%par_LE_cut_off_par.val)
                    print()

            #update gamma_3p and gamma_cut from new B
            gamma_3p_Sync,completed=find_gamma_Synch(self.nu_p_S_obs,self.rest_frame,B_par.val,self.beaming,z_par.val)
            if turn_over_par is not None:
                _t,completed=find_turn_over(jet_model,self.emitters_distr_spectral_type,gamma_3p_Sync)
                turn_over_par.set(val=set_lin_log_val(turn_over_par,_t))

                if silent is False:
                    print("---> setting par type low-energy-cut-off, corresponding to par %s"%(turn_over_par.name))
                    print("---> task completed", completed)
                    print("---> task completed ", completed)
                    print("---> using gamma_3p_Sync=",gamma_3p_Sync)
                    print("---> to %e"%turn_over_par.val)
                    print()
                
            
            #update gmax for New B
            gmax,completed=find_HE_cut_off(self.emitters_distr_spectral_type,self.S_nu_max,self.rest_frame,B_par.val,self.beaming,z_par.val)
            if silent is False:
                print("---> gamma_max=%e from nu_max_Sync= %e, using B=%e"%(gmax,self.S_nu_max,B_par.val))
                print("---> task completed", completed)
            if HE_cut_off_par is not None:
                HE_cut_off_par.set(val=set_lin_log_val(HE_cut_off_par,gmax))

                if silent is False:
                    print("---> setting par type high-energy-cut-off, corresponding to par %s"%(HE_cut_off_par.name))
                    print("---> set to %e"%HE_cut_off_par.val)
                    print()
            
            #update N for New B
            if N_par is not None:
                completed = rescale_Ne(jet_model, self.nuFnu_p_S_obs, self.nu_p_S_obs, self.rest_frame)
                if silent is False:
                    print("---> setting par type emitters_density, corresponding to par %s"%(N_par.name))
                    print("---> to N=%e" % (N_par.val))
                    print("---> task completed",completed)
                    print()
                #N_par.set(val=N)

        #Improve R according to CD
        if silent is False:
            print ("---> setting R from Compton Dominance (CD)")

        R_start=R_tvar
        if EC_componets_list  is None:
            (R_from_CD,failed),completed=constr_R_from_CD(jet_model,self.nuFnu_p_S_obs,self.nu_p_S_obs,self.nuFnu_p_IC_obs,self.nu_p_IC_obs,self.rest_frame,R_tvar,params_grid_size,silent=silent)
        else:
            (R_from_CD,failed),completed=constr_R_from_CD(jet_model,self.nuFnu_p_S_obs,self.nu_p_S_obs,self.nuFnu_p_IC_obs,self.nu_p_IC_obs,self.rest_frame,R_tvar,params_grid_size,EC=True,silent=silent)

        if failed is False and completed is True:
            #
            if R_par is not None:
                R_par.set(val=set_lin_log_val(R_par,R_from_CD))
                if silent is False:
                    print("---> setting par type region_size, corresponding to par %s"%(R_par.name))
                    print("---> set to %e"%R_par.val)
                    print("---> task completed", completed)
                #if silent is False:
                #    print("---> ",R_par.get_description())
                #    print()

                if N_par is not None:
                    completed=rescale_Ne(jet_model,self.nuFnu_p_S_obs,self.nu_p_S_obs,self.rest_frame)
                    if silent is False:
                        print("---> updating setting par type emitters_density, corresponding to par %s"%(N_par.name))
                        print("---> set to %e"%N_par.val)
                        print("---> task completed", completed)
        else:
            R_par.set(val=set_lin_log_val(R_par,R_start))

            if silent is False:
                print("---> constrain failed, R unchanged: ")
                print("---> set to %e"%R_par.val)
                print()
            
         
        #Check tau_gamma_gamma and t_var
        #check_gamma_tansp(jet_model,self.beaming,sp.logspace(21,28,10),self.rest_frame)
        if silent is False:
            print("---> t_var (days)", check_t_var(R_par.val_lin,self.beaming,z_par.val)/86400.)
    
                
        #jet_model.set_flag('obs_constrain_final')

        if silent is False:
            print()
        
            print("show pars")

        jet_model.show_pars()
        
        #sets to initial values
        
        #if self.nu_p_IC_obs is not None:
        jet_model.set_IC_nu_size(IC_nu_size_initial)
        jet_model.set_seed_nu_size(nu_seed_size_initial)

        if silent is False:
           print("eval_model")

        jet_model.eval(fill_SED=True)
        
        #jet_model.set_flag(flag_initial)
        #jet_model.set_path(path_initial)
        
        
#        if self.SEDShape.L_host!=None:
#           
#            SED_model=FitModel(jet=jet_model,template=self.SEDShape.host)
#       
#        else:
#        
#            SED_model=FitModel(jet=jet_model,template=self.SEDShape.host)

            
        return jet_model
        

def run_task(func):

    def func_wrapper(*args, **kwargs):
        completed=True

        try:
            return func(*args, **kwargs),completed
        except Exception as e:
            print('Exception',e)
            return None, completed
        print("-" * 20)

    return func_wrapper


def check_t_var(R,beaming,z):
    return R*(1+z)/(BlazarSED.vluce_cm*beaming)

@run_task
def set_gmin_from_nu_cut_IR(nu_cut_IR,rest_frame,B,beaming,z):
        if nu_cut_IR is not None:
            _v, completed= find_gamma_Synch (nu_cut_IR,rest_frame,B,beaming,z)
            return _v
        else:
            print ("--> !! No nu_cut_IR provided set gmin to 1")
            return 1
           
        
        
        
@run_task
def find_turn_over(jet,distr_e,gamma_3p):
    
    if distr_e=='lppl' or distr_e=='lp':
        r=jet.get_par_by_type('spectral_curvature').val
        s=jet.get_par_by_type('LE_spectral_slope').val
        _v, completed =find_gamma0(r,s,gamma_3p)
        return _v

    elif distr_e=='lpep':
        r=jet.get_par_by_type('spectral_curvature').val
        return gamma_3p/10**(1.5/r)

    #!! TO FIX
    elif distr_e=='plc':
        return gamma_3p*2

    
    elif distr_e=='bkn':
        "!!!!!!!!! MUST FIX ACCORDING P1 and P2!!!!!!"
        return gamma_3p


@run_task
def find_HE_cut_off(distr_e,nu_S_max,rest_frame,B,beaming,z):
    gamma_max_Sync,completed=find_gamma_Synch(nu_S_max,rest_frame,B ,beaming,z)
    
    if distr_e=='lppl' or distr_e=='lp' or distr_e=='lpep':
        return gamma_max_Sync
    
    elif  distr_e=='bkn' or distr_e=='plc' or distr_e=='pl':
        return gamma_max_Sync
    raise RuntimeError ('no gmax found!',distr_e)

@run_task
def find_gamma0(r,s,gamma_3p):
    """returns the value of gamma_0  for
    a log_par+pl distribution
    
    Args:
        r: curvature
        s: spectral index in the PL branch
        gamma_3p: peak value requested for n(gamma)gamma^3

    Returns:
        
        
    """
    if (r!=0.0):
        c=(3-s)/(2*r)
        gamma_0= gamma_3p/(pow(10,c))
    else:
        gamma_0= gamma_3p

    if gamma_0<1.0:
        gamma_0=1.0
    
    return gamma_0



@run_task
def find_B_from_nu_p_S(nu_p_S,gamma_3p,rest_frame,beaming,z):
    """returns B according to Ep_S and gamma_3p


        Args:
            nu_p_S:  Synchrotron peack frequency
            gamma_3p: peak value of n(gamma)gamma^3
            

            re_eval: def==True, set the flag to re_evaluate find_gamma_3p_Synch, after updating B

        Returns:
            B
    """
    nu_p_blob=convert_nu_to_blob(nu_p_S,rest_frame,beaming,z)

    #first tryal for B
    B= (nu_p_blob)/(gamma_3p*gamma_3p*3.7E6 )
    
    return B    
    
@run_task
def  find_gamma_Synch (nu_S,rest_frame,B,beaming,z):
    """returns the value of gamma corresponding to the  Synch freq
    
        
        Args:
            B: magnetic field
            nu_S:  Synchrotron  frequency
            rest_frame: rest frame cooresponding to nu_p_S
            z: redshift
            beamign: beaming factor
            
        Returns:
            gamma
            
    """
    nu_S_blob=convert_nu_to_blob(nu_S,rest_frame,beaming,z)
     
    return m.sqrt(nu_S_blob/(3.7E6*B))


@run_task
def find_gamma_3p_SSC(nu_p_S,nu_p_IC,rest_frame,gamma_3p_Sync,beaming,z,nu_p_EC_seed_field=None,silent=False):
    """returns the value of gamma_3p from nu_p_S/nu_p_IC


        Args:
            nu_p_S :  Synchrotron peack frequency
            nu_p_IC:  IC peack frequency
            rest_frame: rest frame cooresponding to peak frequencies

                    
        Returns:
            gamma_3p_SSC
    """
    #print "*** find_gamma_3p_SSC ***"
    #print "nu_p_S",nu_p_S
    #print "nu+p_IC",nu_p_IC
    #print "beaming",beaming
    #print "z",z
    
    

    
    #print "--> gamma_TH, gamma_KN_Ep_C",g_TH,gamma_KN
    if nu_p_EC_seed_field is None:
        g_TH=(4./3)*m.sqrt(nu_p_IC/nu_p_S)
        nu_p_seed_blob=convert_nu_to_blob(nu_p_S,rest_frame,beaming,z)
       
    else:
        nu_p_seed_blob=nu_p_EC_seed_field/(1+z)*beaming
        g_TH=(4./3)*m.sqrt(nu_p_IC/nu_p_seed_blob)
    #print"--> comp_fac,gamma_TH", comp_factor,g_TH,
    comp_factor,completed=get_Comp_factor(gamma_3p_Sync,nu_p_seed_blob)
    if silent is False:
        print ("---> nu_p_seed_blob=%e"%nu_p_seed_blob)

    gamma_3p_SSC_TH=g_TH
    #print "--> gamma_3p_SSC_TH",gamma_3p_SSC_TH,g_TH
    if silent is False:
        print ("---> COMPTON FACTOR=%e"%comp_factor,gamma_3p_SSC_TH)
    
    if comp_factor>0.01:
        #print "--> correction", 0.2*pow(comp_factor,-0.45)
        gamma_3p_SSC=g_TH/(0.2*pow(comp_factor,-0.45))
    else:
        gamma_3p_SSC=gamma_3p_SSC_TH
    
    #   print "--> gamma_3p_SSC=",gamma_3p_SSC
    return gamma_3p_SSC


@run_task
def get_Comp_factor(gamma,nu_p_S_blob):
    """returns the compton factor = nu_blob_seed_Synch*hplanck/(mec2)
        
        Args:
            gamma: gamma of the up-scattering electrons
            u_p_S_blob: Synchrotron peack frequency in the blob rest frame
           
        Returns:
         compton factor
    """
    # print "--> Ep_S_blob ",nu_p_S_blob
    return gamma*nu_p_S_blob*BlazarSED.HPLANCK/(BlazarSED.MEC2)




@run_task
def find_s(class_obj,nu_p_S_obs,S_LE_slope,indices,silent=False):
    """Find the index of the low-energy PL branch of n(gamma), from PL fit over various instrument bands

    Args:
        class_obj: object  class
        indices_array

    Returns:
        s_Planck, s_X, s_Fermi, s

    """
    
    #1 get S index from nu_p_S_obs
    #2 get IC index from nu_p_IC_obs

    
    if (class_obj=='LSP' or class_obj=='ISP' or class_obj=='HSP'):
        
        s_radio_mm=None
        if indices.get_by_name('radio_mm')is not None:
            if indices.get_by_name('radio_mm').val is not None:
                radio_mm_index=indices.get_by_name('radio_mm').val.spectral
                s_radio_mm=(1.0-2*radio_mm_index)
                if silent is False:
                    print ("---> s_radio_mm",radio_mm_index,s_radio_mm)

                
        
        s_X=None
        if indices.get_by_name('X')is not None:
            if indices.get_by_name('X').val is not None:
                X_index=indices.get_by_name('X').val.spectral
                s_X=(1.0-2*X_index)
                if silent is False:
                    print ("---> s_X",s_X)
    


      
        s_Fermi=None
        if indices.get_by_name('Fermi') is not None:
            if indices.get_by_name('Fermi').val is not None:
                Fermi_index=indices.get_by_name('Fermi').val.spectral
                s_Fermi=(0.4+ Fermi_index*-1*1.60)
                if silent is False:
                   print ("---> s_Fermi",s_Fermi)
    
        
        s_UV_X=None
        if indices.get_by_name('UV_X') is not None:
            if indices.get_by_name('UV_X').val is not None:
                UV_X_index=indices.get_by_name('UV_X').val.spectral
                s_UV_X=(1+ UV_X_index*-1*2.0)
                if silent is False:
                    print ("---> s_UV_X",s_UV_X)
        
        s_Opt_UV=None
        if indices.get_by_name('Opt_UV') is not None:
            if indices.get_by_name('Opt_UV').val is not None:
                Opt_UV_index=indices.get_by_name('Opt_UV').val.spectral
                s_Opt_UV=(1 -1.0*Opt_UV_index*2.0)
                if silent is False:
                    print ("---> s_Opt_UV",Opt_UV_index,s_Opt_UV)
        
        if class_obj=='LSP' :
            #print s_X
            if s_X is not None and s_X-3.0<0:
                #if X_index+1.0>0.2:
                s=s_X
                if silent is False:
                    print ("---> s from X_index",s)
            elif s_radio_mm is not None and s_radio_mm+2>0.0:
                s=s_radio_mm
                if silent is False:
                    print ("---> s from radio_mm_index")
            else:   
                s=2.0
        
        if class_obj=='ISP':
            if s_radio_mm is not None and s_radio_mm+2>0.2:
                s=s_radio_mm
                if silent is False:
                    print ("---> s from radio_mm_index")
            else:   
                s=2.0

        s_SED_shape_photon=S_LE_slope-2.0
        s_SED_shape=-2.0*s_SED_shape_photon-1.0
        if silent is False:
            print ("---> s from synch log-log fit",s_SED_shape)

        if class_obj=='HSP':
           
            s_UV=None
            
            if s_Opt_UV is None and s_UV_X is not None:
                s_UV=s_UV_X
            elif s_Opt_UV is not None and s_UV_X is None:
                s_UV=s_Opt_UV
            elif s_Opt_UV is not None and s_UV_X is not None:
                s_UV=s_UV_X
                
            if s_Fermi is not None and s_UV is not None:
                s=(s_Fermi +s_UV )/(2.0)
                if silent is False:
                    print ("---> s from (s_Fermi + s_UV)/2")
            elif s_UV is not None :
                s=s_UV
                if silent is False:
                    print ("---> s from  s_UV")

            elif s_Fermi is not None :
                s=s_Fermi
                if silent is False:
                    print ("---> s from Fermi")

            
            elif s_radio_mm is not None :
                s=s_radio_mm
                if silent is False:
                    print ("---> s from radio_mm")

            else:
                s=2.0
    if silent is False:
        print ("---> power-law index s, class obj=%s s chosen is %f"%(class_obj,s))
    res = (s_radio_mm, s_X, s_Fermi,s)
    return res


@run_task
def find_s1(class_obj,indices):
    #print "---> !!! fake function, still to develop"
    val=3.5
    #print ("---> set s1 to %f"%val)
    return val

@run_task
def rescale_Ne(jet,S_p,nu_p,rest_frame):
    """Rescales N.blob to get the wanted Lp_S_blob
    """

    if rest_frame=='obs':
        jet.set_N_from_nuFnu(nuFnu_obs=S_p,nu_obs=nu_p)
    else:
        jet.set_N_from_nuLnu(nuLnu_src=S_p, nu_src=nu_p)

    return


@run_task
def get_U_Sync_from_Ph(jet,re_eval_sync=False):
    """returns U_synch by integrating Synch photon spectrum

    Args:
        blob: SED module class
        temp_ev: SED module class

        re_eval_sync: def==False, re_eval Synch spectrum and then get U_synch
    Returns:
        U_synch

    """
    if re_eval_sync==True:
        comp_old=jet.blob.do_SSC
        jet.blob.do_SSC=0
        jet.eval()
        jet.blob.do_SSC=comp_old
    
    return BlazarSED.Uph_Sync(jet.blob)

def check_boundaries(val,val_min,val_max,val_name,silent=False):
    failed=False
    if val<val_min or val>val_max:
        failed=True
        if silent is False:
            print ("---> %s=%e, out of boundaries %e %e, rejected"%(val_name,val,val_min,val_max))

        if val<val_min:
            val=val_min
        elif val>val_max:
            val=val_max

        if silent is False:
            print ("     Best %s not found, (temporary set to %e)"%(val_name,val))
    else:
        if silent is False:
            print ("     Best %s=%e"%(val_name,val))

    return val,failed

@run_task
def get_R_tvar(beaming,t_var_sec,z):
    return BlazarSED.vluce_cm*beaming*t_var_sec/(1+z)

def set_lin_log_val(p,v):
    if p.islog is True:
        v = m.log10(v)
    return v

@run_task
def constr_R_from_CD(jet,nuFnu_p_S,nu_p_S,nuFnu_p_IC,nu_p_IC,rest_frame,R_tvar,params_grid_size,EC=False,silent=False):
    #################################
    # This function must not        #
    # change jet.blob attributes    #
    # all the changed values must   #
    # be set back to their original #
    # values                        #
    #################################
    R_initial=jet.get_par_by_name('R').val_lin
    N_initial=jet.get_par_by_name('N').val

    CD_model_log=[]
    CD_obs=nuFnu_p_IC/nuFnu_p_S
    R_min=R_tvar/1000
    R_min=1E13
    R_max=R_tvar*2
    R_grid=np.logspace(np.log10(R_min), np.log10(R_max), params_grid_size)
    for R in R_grid:

        jet.set_par('R',val=set_lin_log_val(jet.get_par_by_name('R'),R))

        completed = rescale_Ne(jet, nuFnu_p_S, nu_p_S, rest_frame)

        jet.eval()

        Lp_S=jet.get_SED_peak(Model_dic.Sync_nuFnu_p_dic[rest_frame])
        if EC==False:
            Lp_IC=jet.get_SED_peak(Model_dic.SSC_nuFnu_p_dic[rest_frame])
        else:
            nu_p_IC,Lp_IC=jet.get_SED_peak(freq_range=[0.01*nu_p_IC,10*nu_p_IC])
        CD=Lp_IC/Lp_S
        CD_model_log.append(np.log10(CD))


    R_grid_log=np.log10(R_grid)
    p=polyfit(CD_model_log,R_grid_log,2)
    best_R=polyval(p, np.log10(CD_obs))
    Best_R=np.power(10., best_R)
    Best_R,failed=check_boundaries(Best_R,R_min,R_max,'R',silent=silent)
    

    jet.set_par('R',val=set_lin_log_val(jet.get_par_by_name('R'),R_initial))
    jet.set_par('N',val=N_initial)
    res = (Best_R,failed)
    return res


@run_task
def constr_B_from_nu_peaks(jet,nu_p_S,nu_p_IC,rest_frame,B_min,B_max,beaming,params_grid_size,EC=False,silent=False):
    #################################
    # This function must not        #
    # change jet.blob attributes    #
    # all the changed values must   #
    # be set back to their original #
    # values                        #
    #################################

    flag_initial=jet.get_flag()
    
    B_initial=jet.get_par_by_name('B').val
    
    jet.set_flag('constr_B')
 
    nu_p_IC_model_log=[]
    
    B_grid=np.logspace(np.log10(B_min),np.log10(B_max),params_grid_size)
    
    #f=open('%s/B_vs_nu_p_IC.dat'%jet.get_path(),'w')
    
    #fin the turn-over variable to update with B
    turn_over_energy=jet.get_par_by_type('turn-over-energy')
    
    if turn_over_energy is not None :
        turn_over_energy_initial=turn_over_energy.val_lin
    
    
    z=jet.get_par_by_type('redshift').val
    
    for B in B_grid:
        
        jet.set_par('B',val=B)
        
        
        gamma_3p,completed=find_gamma_Synch(nu_p_S,rest_frame,B,beaming,z)

        #upda-turn-over variable
        if turn_over_energy is not None :
            turn_over_energy.set(val=set_lin_log_val(turn_over_energy,gamma_3p))

        
        jet.eval()   
        
        if EC==False:
            nu_p_model=jet.get_SED_peak(Model_dic.SSC_nu_p_dic[rest_frame])
        else:
             nu_p_model,nuFnu_p_model=jet.get_SED_peak(freq_range=[0.01*nu_p_IC,10*nu_p_IC])
        
        nu_p_IC_model_log.append(np.log10(nu_p_model))

        #print(n.log10(B),n.log10(nu_p_model),file=f)
        #print n.log10(B),n.log10(nu_p_model)

    #f.close()
    
    B_grid_log=np.log10(B_grid)
    
     
    p=polyfit(nu_p_IC_model_log,B_grid_log,2)    
    
    Best_B=polyval(p, np.log10(nu_p_IC))
    Best_B = np.power(10., Best_B)
    Best_B,failed  = check_boundaries(Best_B, B_min, B_max, 'B',silent=silent)
    
    
    #REST CHANGED VALUES
    jet.set_flag(flag_initial)
    jet.set_par('B',val=B_initial)

    if turn_over_energy is not None :
        turn_over_energy.set(val=set_lin_log_val(turn_over_energy,turn_over_energy_initial))

    res= (Best_B,failed)
    return res

#def check_gamma_transp(jet, beaming_val, nu_IC_data, rest_frame):
#     """retrun tau_gamma_gamma for a given  IC frequency
#
#     Args:
#         blob:
#         temap_ev:
#         nu_IC_data: data freq of the IC component to check fro tau_gamma_gamma
#
#     Returns:
#         tau_gamma_gamma
#
#     """
#     #################################
#     # This function must not        #
#     # change jet.blob attributes    #
#     # all the changed values must   #
#     # be set back to their original #
#     # values                        #
#     #################################
#
#     flag_initial=jet.get_flag()
#
#     jet.set_flag('tau_gamma_gamma')
#
#     IC_initial=jet.get_IC_mode()
#
#     jet.set_IC_mode('off')
#
#
#     #nu_IC_blob=convert_nu_to_blob(nu_IC_data,rest_frame,blob.beam_obj,blob.z_cosm)
#     #target_nu_blob = SED.MEC2 * SED.MEC2 / (SED.HPLANCK * SED.HPLANCK * nu_IC_blob)
#     target_nu_obs=BlazarSED.MEC2 * BlazarSED.MEC2 / (BlazarSED.HPLANCK * BlazarSED.HPLANCK * nu_IC_data)
#     #print nu_IC_blob,target_nu,target_nu*nu_IC_blob
#
#
#     jet.eval()
#
#
#     #nu_obs=sp.array([])
#     #nuFnu_obs=sp.array([])
#     nu_obs,nuFnu_obs=jet.spectral_components.Sync.get_SED_points(log_log=True,name='Sync')
#
#     #for i in range(BlazarSED.GetNuIntMaxSynch()):
#     #
#     #     if BlazarSED.GetSEDSynch(i)>0:
#     #         nu_obs=sp.append(nu_obs,sp.log10(BlazarSED.GetNuObsSynch(i)))
#     #        nuFnu_obs = sp.append(nuFnu_obs, sp.log10(BlazarSED.GetSEDSynch(i)))
#
#
#     nuFnu_target = sp.power(10.0,sp.interp(sp.log10(target_nu_obs),nu_obs,nuFnu_obs))
#
#     if sp.shape(target_nu_obs)==():
#         target_nu_obs=sp.array([target_nu_obs])
#
#         nuFnu_target=sp.array([nuFnu_target])
#
#         nu_IC_data=sp.array([nu_IC_data ])
#
#     for i in range(len(target_nu_obs)):
#         #print nuFnu_target,blob.beam_obj,blob.z_cosm,blob.dist
#
#
#
#         z_val=jet.get_par_by_type('redshift').val
#
#         DL_cm_val=jet.get_DL_cm()
#
#         nuLnu_target_blob= BlazarSED.nuFnu_obs_to_nuLnu_blob(nuFnu_target[i],beaming_val,z_val,DL_cm_val)
#
#         R_val=jet.get_par_by_type('region_size').val_lin
#
#         tau_gamma_gamma=BlazarSED.SIGTH/5*nuLnu_target_blob/(4*sp.pi * BlazarSED.vluce_cm * R_val * BlazarSED.MEC2 )
#
#         print ("     -->target comoving freq=%e"%target_nu_obs[i], "observed freq=%e"%nu_IC_data[i], "tau_gamma=%e"%(tau_gamma_gamma))
#
#
#     jet.set_IC_mode(IC_initial)
#     jet.set_flag(flag_initial)


import pytest
from .base_class import TestBase

class TestDependingParameters(TestBase):

    def integration_suite(self,plot=False):
        self._all(plot=plot)

    def test_dep_par(self,plot=False):
        from jetset.jet_emitters import EmittersDistribution
        import numpy as np

        def distr_func_bkn(gamma_break, gamma, s1, s2):
            return np.power(gamma, -s1) * (1. + (gamma / gamma_break)) ** (-(s2 - s1))

        n_e_bkn = EmittersDistribution('bkn', spectral_type='bkn')
        n_e_bkn.add_par('gamma_break', par_type='turn-over-energy', val=1E3, vmin=1., vmax=None, unit='lorentz-factor')
        n_e_bkn.add_par('s1', par_type='LE_spectral_slope', val=2.5, vmin=-10., vmax=10, unit='')
        n_e_bkn.add_par('s2', par_type='LE_spectral_slope', val=3.2, vmin=-10., vmax=10, unit='')
        n_e_bkn.set_distr_func(distr_func_bkn)
        n_e_bkn.parameters.show_pars()
        n_e_bkn.parameters.s1.val = 2.0
        n_e_bkn.parameters.s2.val = 3.5
        n_e_bkn.update()
        n_e_bkn.parameters.show_pars()

        from jetset.jet_model import Jet

        j = Jet(emitters_distribution=n_e_bkn)

        # def par_func(s1):
        #    return s1+1
        j.make_dependent_par(par='s2', depends_on=['s1'], par_expr='s1+1')
        print('here')
        j.parameters.s1.val = 3
        print('done')
        np.testing.assert_allclose(j.parameters.s2.val, j.parameters.s1.val + 1)
        j.save_model('jet.pkl')
        new_jet = Jet.load_model('jet.pkl')
        print('here')
        new_jet.parameters.s1.val = 2
        print('done')

        np.testing.assert_allclose(new_jet.parameters.s2.val, new_jet.parameters.s1.val + 1)
        j.eval()
        new_jet.show_model()


    def test_dep_par_log_values(self,plot=False):
        from jetset.jet_emitters import EmittersDistribution
        import numpy as np

        def distr_func_bkn(gamma_break, gamma, s1, s2):
            return np.power(gamma, -s1) * (1. + (gamma / gamma_break)) ** (-(s2 - s1))

        n_e_bkn = EmittersDistribution('bkn', spectral_type='bkn')
        n_e_bkn.add_par('gamma_break', par_type='turn-over-energy', val=3, vmin=0.1, vmax=None, unit='lorentz-factor',log=True)
        n_e_bkn.add_par('s1', par_type='LE_spectral_slope', val=2.5, vmin=-10., vmax=10, unit='')
        n_e_bkn.add_par('s2', par_type='LE_spectral_slope', val=3.2, vmin=-10., vmax=10, unit='')
        n_e_bkn.set_distr_func(distr_func_bkn)
        n_e_bkn.parameters.show_pars()
        n_e_bkn.parameters.s1.val = 2.0
        n_e_bkn.parameters.s2.val = 3.5
        n_e_bkn.update()
        n_e_bkn.parameters.show_pars()

        from jetset.jet_model import Jet

        j = Jet(emitters_distribution=n_e_bkn)

        # def par_func(s1):
        #    return s1+1

        j.make_dependent_par(par='gamma_break', depends_on=['gmax'], par_expr='gmax / 10')
        j.parameters.gmax.val = 1E5
        print('par',j.parameters.gamma_break.val, j.parameters.gmax.val / 10)
        np.testing.assert_allclose(j.parameters.gamma_break.val, np.log10(j.parameters.gmax.val/ 10))
        j.save_model('jet.pkl')
        new_jet = Jet.load_model('jet.pkl')
        new_jet.parameters.gmax.val = 1E4
        np.testing.assert_allclose(new_jet.parameters.gamma_break.val, np.log10(new_jet.parameters.gmax.val/ 10))
        j.eval()
        new_jet.show_model()

        j = Jet(emitters_distribution='lppl',emitters_distribution_log_values=True)
        j.show_pars()
        j.parameters.gmax.val = 5
        j.make_dependent_par(par='gamma0_log_parab', depends_on=['gmax'], par_expr='gmax / 10')
        j.show_pars()  
        
        print('par',j.parameters.gamma0_log_parab.val, j.parameters.gmax.val - 1)
        np.testing.assert_allclose(j.parameters.gamma0_log_parab.val, j.parameters.gmax.val - 1)
        print(j.parameters.gamma0_log_parab.val, j.parameters.gmax.val -1)
        j.save_model('jet.pkl')
        new_jet = Jet.load_model('jet.pkl')
        new_jet.parameters.gmax.val = 4
        np.testing.assert_allclose(new_jet.parameters.gamma0_log_parab.val, new_jet.parameters.gmax.val -1)
        j.eval()
        new_jet.show_model()


    def test_dep_par_jet(self,plot=False):
        import numpy as np

        from jetset.jet_model import Jet
        j = Jet(emitters_distribution='plc')
        j.add_user_par(name='B0',units='G',val=1E-5,val_min=0,val_max=None)
        j.make_dependent_par(par='B',depends_on=['B0'],par_expr='B0*5')
        np.testing.assert_allclose(j.parameters.B.val, j.parameters.B0.val*5)

        j.save_model('test.pkl')
        new_j=Jet.load_model('test.pkl')
        new_j.parameters.B0.val=1
        np.testing.assert_allclose(new_j.parameters.B.val, new_j.parameters.B0.val*5)


    def test_dep_par_composite_model(self,plot=False):
        import numpy as np
        from jetset.model_manager import FitModel
        from jetset.jet_model import Jet

        jet = Jet(emitters_distribution='plc')
        fit_model = FitModel(jet=jet, name='SSC-best-fit-lsb', template=None)
        fit_model.jet_leptonic.parameters.beam_obj.fit_range = [5, 50]
        fit_model.jet_leptonic.parameters.R_H.val = 5E17
        fit_model.jet_leptonic.parameters.R_H.frozen = False
        fit_model.jet_leptonic.parameters.R_H.fit_range = [1E15, 1E19]
        fit_model.jet_leptonic.parameters.R.fit_range = [10 ** 15.5, 10 ** 17.5]
        fit_model.jet_leptonic.add_user_par(name='B0', units='G', val=1E3, val_min=0, val_max=None)
        fit_model.jet_leptonic.add_user_par(name='R0', units='cm', val=5E13, val_min=0, val_max=None)
        fit_model.jet_leptonic.add_user_par(name='m_B', val=1, val_min=1, val_max=2)

        fit_model.jet_leptonic.parameters.R0.frozen = True
        fit_model.jet_leptonic.parameters.B0.frozen = True

        def par_func(R0,B0,R_H,m_B): 
            return B0*np.power((R0/R_H),m_B)

        fit_model.jet_leptonic.make_dependent_par(par='B', depends_on=['B0', 'R0', 'R_H','m_B'], par_expr=par_func)

        B0=fit_model.jet_leptonic.parameters.B0.val
        R0 = fit_model.jet_leptonic.parameters.R0.val
        R_H = fit_model.jet_leptonic.parameters.R_H.val
        m_B= fit_model.jet_leptonic.parameters.m_B.val

        np.testing.assert_allclose(fit_model.jet_leptonic.parameters.B.val, par_func(B0,R0,R_H,m_B))

        fit_model.save_model('test_composite.pkl')
        fit_model.parameters._build_par_table()
        t=fit_model.parameters._par_table
        t.sort(t.colnames)
        
        for new_fit_model in [FitModel.load_model('test_composite.pkl'),fit_model.clone()]:
            
            new_fit_model.parameters._build_par_table()
            _t=new_fit_model.parameters._par_table
            _t.sort(_t.colnames)

            assert(all([_t[i]['frozen']==t[i]['frozen'] for i in range(len(_t))]))
            assert(all([_t[i]['val']==t[i]['val'] for i in range(len(_t))]))

            new_fit_model.jet_leptonic.parameters.B0.val=1E4

            B0 = new_fit_model.jet_leptonic.parameters.B0.val
            R0 = new_fit_model.jet_leptonic.parameters.R0.val
            R_H = new_fit_model.jet_leptonic.parameters.R_H.val
            
            np.testing.assert_allclose(new_fit_model.jet_leptonic.parameters.B.val, par_func(B0,R0,R_H,m_B))

            


__author__ = "Andrea Tramacere"

from astropy.units import  Unit as u
from astropy import cosmology 

__all__=['Cosmo']


class  Cosmo (object):


    def __init__(self,astropy_cosmo=None,DL_cm=None):

        _c = None

        if DL_cm is not None and astropy_cosmo is not None:
            raise  RuntimeError('Either you provide an astropy cosmology objet, or luminosity distance in cm, or nothing')

        elif astropy_cosmo is None and DL_cm is None:
           
            _c = cosmology.Planck13

        elif astropy_cosmo is not None and DL_cm is None:

            _c = astropy_cosmo

        elif astropy_cosmo is None and DL_cm is not None:

            _c=None
            print("using cosmo without z and only DL, should be used only for galactic objects!!")
            print("z will be fixed to zero")
        else:
            raise RuntimeError('Either you provide an astropy comsology objet, or luminosity distance in cm, or nothing')


        self._c = _c
        self._DL_cm = DL_cm

    def __repr__(self):
        if self._c is not None:
            s= '%s' % self._c
        elif self._DL_cm is not None:
            x=self._DL_cm*u('cm')

            s= 'cosmology is not defined, the luminosity distance has been set to %s'%x
        else:
            s='you have to define either an astropy comoslogy model or set a luminosity distance in cm'

        return s

    def get_DL_cm(self,z=None):
        if self._c is not None:
            #THIS IS FIXING THE ERROR WITH PICKLED COSMO
            #TODO open issue on astropy!
            try:
                _d= self._c.luminosity_distance( z ).to('cm').value
            except:
                _d = self._c.luminosity_distance(z)
                _d = _d.value*u(str(_d.unit))
                _d = _d.to('cm').value
        else:
            _d = self._DL_cm.value

        return _d
    
    







__author__ = "Andrea Tramacere"

import  numpy as np
import warnings
from astropy.table import Table,vstack,MaskedColumn
from astropy import  units as u
from .utils import clean_var_name
from functools import wraps
import ast
import inspect


__all__=['ModelParameter','ModelParameterArray','Value']


def is_notebook():
    try:
        from IPython import get_ipython
        if "IPKernelApp" not in get_ipython().config:  # pragma: no cover
            raise ImportError("console")
            return False
    except:
        return False
    else:  # pragma: no cover
        return True

def _show_table(t):
    if is_notebook():
        try:
            from IPython.display import display
            display(t.show_in_notebook(show_row_index=False, display_length=100))
        except:
            try:
                from IPython.display import display
                display(t)
            except:
                t.pprint_all()
    else:
        t.pprint_all()

class Value(object):

    def __init__(self,val,units,islog=False):
        self.val=val
        self.islog=islog
        self.units=units

    def __repr__(self):
        return '%s'%self._val

    @property
    def islog(self):
        return  self._islog

    @islog.setter
    def islog(self,val):
        self._islog=val

    @property
    def val(self):
        return self._val

    @val.setter
    def val(self, val):
        self._val = val

    @property
    def lin(self):
        if self._val is None:
            return None
        if self._islog is True:
            return 10**self._val
        else:
            return self._val

    @property
    def log(self):
        if self._val is None:
            return None
        if self._islog is True:
            return self._val
        else:
            return np.log10(self._val)

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, p_unit,verbose=False):
        try:
            self._units = u.Unit(p_unit)
            #print(units,type(self._units))
        except Exception as e:
            if verbose is True:
                print('units not valid for astropy units ',p_unit)

            if hasattr(self,'_units'):
                if '*' not in str(self._units):
                    self._units = p_unit + '*'
            else:
                if '*' not in p_unit:
                    self._units = p_unit + '*'
                else:
                    self._units = p_unit


class ModelParameter(object):
    """
    This class is the base class for models parameters. The following keywords 
    arguments can be passed to the constructor
    
    Parameters
    ----------      
    name          :  (str)    parameter name (default='unk')
    val           :  (float)  parameter value 
    par_type      :  (str)    parameter type (default='unk')
    units         :  (str)    units of the parameter, (default='No')
    val_min       :  (float)  minimum physical value 
    val_max       :  (float)  maximum physical value
    val_start     :  (float)  starting value
    val_last_call :  (float)  last call value
    fit_range_min :  (float)  minimum boundary value for the fit
    fit_range_max :  (float)  maximum boundary value for the fit  
    frozen        :  (bool)   boolean flag for frozen parameter (default=False)
    log           :  (bool)   boolean flag for log-scale value (default=False)
                    
    Attributes
    ----------
    name          :  (str)         parameter name (default='unk')                      
    val           :  (float)       parameter value                                    
    par_type      :  (str)         parameter type (default='unk')                     
    units         :  (str)         units of the parameter, (default='No')             
    val_min       :  (float)       minimum physical value                             
    val_max       :  (float)       maximum physical value                             
    val_start     :  (float)       starting value                                     
    val_last_call :  (float)       last call value                                    
    fit_range_min :  (float)       minimum boundary value for the fit                 
    fit_range_max :  (float)       maximum boundary value for the fit                 
    frozen        :  (bool)        boolean flag for frozen parameter (default=False)  
    log           :  (bool)        boolean flag for log-scale value (default=False)    
 
    """
    
    
    
    def __init__(self, **keywords):


        self.allowed_keywords={'name':None}
        self.allowed_keywords['val']=None
        self.allowed_keywords['par_type']=None
        self.allowed_keywords['allowed_par_types']=None
        self.allowed_keywords['units']='No'
        self.allowed_keywords['val_min']=None
        self.allowed_keywords['val_max']=None
        self.allowed_keywords['val_start']=None
        self.allowed_keywords['val_last_call']=None
        self.allowed_keywords['fit_range_min']=None
        self.allowed_keywords['fit_range_max']=None
        self.allowed_keywords['fit_range']=None
        self.allowed_keywords['best_fit_val']=None
        self.allowed_keywords['best_fit_err']=None
        self.allowed_keywords['frozen']=False
        self.allowed_keywords['log']=False
        self.allowed_keywords['allowed_values'] = None
        self.allowed_keywords['_is_dependent'] = False
        self.allowed_keywords['_depending_pars'] = []
        self.allowed_keywords['_func'] = None
        self.allowed_keywords['_master_pars'] = []
        self.allowed_keywords['_root_par'] = None
        self.allowed_keywords['_linked'] = False
        self.allowed_keywords['_linked_root_model'] = None
        self.allowed_keywords['_master_par_list'] = []
        self.allowed_keywords['_depending_par_expr'] = None
        self._linked = False
        self._linked_root_model = None
        self._root_par = None
        _v = None
        _l = False
        _units= None
        self.par_type=None

        if '_is_dependent' in keywords.keys():
            self._is_dependent =  keywords['_is_dependent']
        else:
            self._is_dependent = self.allowed_keywords['_is_dependent']

        if '_depending_pars' in keywords.keys():
            self._depending_pars = keywords['_depending_pars']
        else:
            self._depending_pars = self.allowed_keywords['_depending_pars']

        if '_func' in keywords.keys():
            self._func = keywords['_func']
        else:
            self._func = self.allowed_keywords['_func']

        if '_master_pars' in keywords.keys():
            self._master_pars = keywords['_master_pars']
        else:
            self._master_pars = self.allowed_keywords['_master_pars']

        if 'val' in keywords.keys():
            _v = keywords['val']

        if 'log' in keywords.keys():
            _l = keywords['log']
        
        if 'units' in keywords.keys():
           _units= keywords['units']

        self._val = Value(val=_v, islog=_l,units=_units)

        for kw in self.allowed_keywords.keys():
            if kw == 'val':
                pass
            elif kw == 'log':
                pass
            elif kw == 'units':
                pass
            elif kw == 'par_type':
                pass
            else:
                setattr(self,kw,self.allowed_keywords[kw])


        self.set(**keywords)




    @property
    def linked(self):
        return self._linked

    @property
    def islog(self):
        return self._val.islog

    @property
    def val(self ):
        return self._val.val

    @property
    def val_lin(self):
        return self._val.lin

    @property
    def val_log(self):
        return self._val.log

    @val.setter
    def val(self,val):
        self.set(val=val)

    @property
    def units(self):
        return self._val.units

    @units.setter
    def units(self,val):
        self._val.units=val

    def to(self, units):
        if isinstance(self.units,u.Unit):
            return self.val*self.units.to(units)
        else:
            message='the units of this parameter:'+self.units+', are not convertible'
            warnings.warn(message)
    @property
    def fit_range(self):

        return [self.fit_range_min,self.fit_range_max]

    @fit_range.setter
    def fit_range(self, fit_range=None):
        if fit_range is None:
            fit_range=[None,None]
        if isinstance(fit_range,tuple):
            pass
        elif isinstance(fit_range,list):
            pass
        else:
            raise RuntimeError('fit_range bust me list or tuple with length=2')
        if len(fit_range)!=2:
            raise RuntimeError('fit_range bust me list or tuple with length=2')

        #self._fit_range = fit_range
        self.fit_range_min=fit_range[0]
        self.fit_range_max=fit_range[1]
        if fit_range[0] is not None and self.val<fit_range[0]:
            raise RuntimeError('par',self.name, 'value',self.val,'< fit range min',fit_range[0])
        if fit_range[1] is not None and self.val>fit_range[1]:
            raise RuntimeError('par',self.name, 'value',self.val,'> fit range max',fit_range[1])

    def reset_dependencies(self):
        self._linked = False
        self._linked_root_model = None
        self._func=None
        self._is_dependent = False
        self._master_pars=[]
        self._depending_pars=[]

    def _add_depending_par(self,par):
        if par not in self._depending_pars:
            self._depending_pars.append(par)

    def _add_master_par(self,par):
        if par not in self._master_pars:
            self._master_pars.append(par)

    #def make_dependent_par(self, master_par, func, root_model=None):

    #    if self == master_par:
    #        raise RuntimeError(" root and linked parameter can't be the same")
    #    self._is_dependent = True
    #    self._func = func
    #    self._master_pars = master_pars
    #    master_par._add_depending_par(self)

    #    self.freeze()
    #    if root_model is not None:
    #        self._linked_root_model = root_model
    #    self.set(val=self._func(self._master_par.val), skip_dep_par_warning=True)


    # def get_default_args(self, par_expr):
    #     #signature = inspect.signature(func)
    #     #d={}
    #     #for k, v in signature.parameters.items():
    #     #    if isinstance(v.default,ModelParameter):
    #     #        d[k]=v.default
    #     #    else:
    #     #        raise RuntimeError('argument',k,'is not valid, should be a model parameter')
    #     #return d
    #     pass

    @property
    def print_par_expr(self):
        if isinstance(self.par_expr,str):
            _par_expr=self.par_expr
        else:
            _par_expr=inspect.getsource(self.par_expr)
        print('==> par', self.name, 'is depending on', [_p.name for _p in self._master_pars],  f'according to expr:   {self.name} =\n{_par_expr}'.format(self.name,_par_expr))
        

    @property
    def par_expr(self):
        return self._depending_par_expr

    @par_expr.setter
    def par_expr(self, expr_string):
        self._depending_par_expr = expr_string

    def _eval_par_func(self):
        #transform par name and value into a local var
        if type(self._depending_par_expr) == str:
            _par_values= [None]*len(self._master_pars)
            for ID, _user_par_ in enumerate(self._master_pars):
                _par_values[ID] = _user_par_.val_lin
                #print('==> _eval_par_func',_user_par_.name,_par_values[ID])
                exec(_user_par_.name + '=_par_values[ID]')
            res = eval(self._depending_par_expr)
        elif callable(self._depending_par_expr) is True:
            _par_values={}
            for ID, _user_par_ in enumerate(self._master_pars):
                _par_values[_user_par_.name] = _user_par_.val_lin
            res=self._depending_par_expr(**_par_values)

        if self.islog is True:
            res=np.log10(res)

        return res
        #return eval(self.par_expr)


    def set(self, *args, skip_dep_par_warning=False, **keywords):
        """
        sets a parameter value checking for physical boundaries
        
        Parameters: keywords of the constructor
         
        """
        keys = keywords.keys()
        if self.immutable is False or skip_dep_par_warning is True:
            pass
        else:
            #warnings.warn('\n\n *** you are trying to set a dependent parameter:%s *** \n'%self.name)
            raise RuntimeError('\n\n *** you are trying to set a dependent parameter:%s *** \n'%self.name)
            #return


        for kw in keys:

            if kw in  self.allowed_keywords.keys() :
                if kw == 'val':

                    if self.allowed_values is not None:

                        if keywords[kw] not in self.allowed_values:
                            raise RuntimeError('parameter  %s' %(self.name), 'the value', keywords[kw] , 'is not in the allowed list',self.allowed_values)

                    self._val.val = keywords[kw]
                    if self._depending_pars is not []:
                        for p in self._depending_pars:
                            #print("==> setting dep par",p.name, 'to',p._func(),'p=',p,'master',self,'name',self.name)
                            p.set(val=p._func(),skip_dep_par_warning=True)

                elif kw == 'log':
                    self._val.islog = keywords[kw]

                elif kw== 'units':
                    self._val.units = keywords[kw]

                elif kw == 'par_type':
                    if self.allowed_par_types is not None:

                        if  keywords[kw] not in self.allowed_par_types:
                            msg = "parameter %s the  type %s is not allowed" % (self.name,keywords[kw]) + "\n please choose among %s" % self.allowed_par_types
                            raise ValueError("%s" % msg)
                    setattr(self, kw, keywords[kw])
                else:
                    setattr(self,kw,keywords[kw])
            else:
                
                print ("wrong keyword=%s, not in%s "%(kw, self.allowed_keywords.keys()))
                
                raise ValueError

        if self.fit_range is not None:
            self.fit_range_max=self.fit_range[1]
            self.fit_range_min=self.fit_range[0]
        
    
        
        #if self.frozen==True:
        #    self.val=self.val_start
        
        self.val_last_call=self.val    
        
        if self.val_min is not None:
            if self.val<self.val_min:
                raise RuntimeError("par=%s  = %e out of boundary=%e"%(self.name,self.val,self.val_min))
         
        if self.val_max is not None:
            if self.val>self.val_max:
                raise RuntimeError("par=%s  = %e out of boundary=%e"%(self.name,self.val,self.val_max))

        if self.val_min is not None and self.fit_range_min is None:
            self.fit_range_min=self.val_min
        
        if self.val_max is not None and self.fit_range_max is None:
            self.fit_range_max=self.val_max
    
    
    
    def get(self,*args ):
        
       
        for arg in args:
            
            if arg in  self.allowed_keywords.keys():
                
                return getattr(self,arg)
                    
            else:
                
                print ("wrong arg=%s, not in%s "%(arg, self.allowed_keywords.keys()))
                
                raise ValueError

    @property
    def immutable(self):
        return self._is_dependent or self._linked

    @property
    def frozen(self):
        return self._frozen

    @frozen.setter
    def frozen(self,v,skip_dep_par_warning=False):
        if self.immutable is True and skip_dep_par_warning is True:
            raise RuntimeError('frozen state of linked/dependent parameter:',self.name , 'can not be changed, please update your script')

        if v not in [True,False]:
            raise RuntimeError('par',self.name,'only True or False are allowed')
        else:
            self._frozen = v



    def freeze(self):
        """
        freezes a parameter
        """
        self.frozen=True

    
        
    def free(self):
        """
        make a parameter free
        """ 
        self.frozen=False



    def get_fit_initial_value(self):
        """
        Gives the initial fit value of the parameter

        Returns
        -------
        val_start : value    
            the parameter initial value
        """
        return self.val_start
    
    def set_fit_initial_value(self,val):
        """
        Sets the initial fit value of the parameter

       
        """
        self.val_start=val
    
    
    def show(self):
        """
        Prints the description of a parameter
        """
        print (self.get_description())
    
    def show_best_fit(self):
        """
        Prints the best-fit description of a parameter
        """
        print (self.get_bestfit_description())
    
    def get_description(self,nofields=False):
        """
        gives the value of each member of the  :class:`ModelParameter` objects, except the best-fit values
 
        Returns
        -------
        descr : (str)
            a string describing all the parameter values, except the best-fit values
        """

        if self.val_min is None:
            val_min='No'
        else:
            val_min=str("%+e"%self.val_min)
            
        if self.val_max is None:
            val_max='No'
        else:
            val_max=str("%+e"%self.val_max)
        
        if self.val is None:
            val='No'
        else:
            val=str("%+e"%self.val)
        
        if nofields==False:
            descr= "name = %-16s  type = %-20s  units = %-16s  val = %s  phys-bounds = [%-13s,%-13s] islog = %s  frozen= %s "%(self.name, self.par_type ,
                                                                                   self.units, val, val_min, val_max,self.islog,self.frozen)
        else:
            descr= " %-16s | %-20s | %-16s | %s | [%-13s,%-13s] | %s | %s"%(self.name, self.par_type ,
                                                                                   self.units, val, val_min, val_max,self.islog,self.frozen)
            
        return descr


    def get_bestfit_description(self,nofields=False):
        """
        gives the value of each member of the  :class:`ModelParameter` objects, suited for  the best-fit values
        
        Returns
        -------
        descr : (str)
            a string describing all the parameter values, suited for  the best-fit values
        """
        
        if self.val_start is None:
            val_start='No'
        else:
            val_start=str("%+e"%self.val_start)
        
        if self.fit_range_min is None:
            fit_range_min='No'
        else:
            fit_range_min=str("%+e"%self.fit_range_min)
            
        if self.fit_range_max is None:
            fit_range_max='No'
        else:
            fit_range_max=str("%+e"%self.fit_range_max)
        
        if self.best_fit_val is None:
            best_fit_val='No'
        else:
            best_fit_val=str("%+e"%self.best_fit_val)

        if hasattr(self,'err_p') and  hasattr(self,'err_m'):
            best_fit_err_m=str("%+e"%self.err_m)
            best_fit_err_p=str("%+e"%self.err_p)
        else:
            best_fit_err_m='#'
            if self.best_fit_err is None:
                best_fit_err_p='No'
            else:
                best_fit_err_p=str("%+e"%self.best_fit_err)
        
        if self.frozen==True:
            best_fit_val='Frozen'
            best_fit_err_p='Frozen'
            best_fit_err_m='Frozen'
        if nofields==False:
            descr= "name = %-16s best-fit val=%-13s  best-fit err p=%-13s best-fit err m=%-13s start-val=%-13s   fit-bounds=[%-13s,%-13s]"%(self.name,best_fit_val,best_fit_err_p,best_fit_err_m,val_start,fit_range_min,fit_range_max)
        else:
            descr= " %-16s | %-13s | %-13s | %-13s | %-13s | [%-13s,%-13s]"%(self.name,best_fit_val,best_fit_err_p,best_fit_err_m,val_start,fit_range_min,fit_range_max)


        return descr

    def identity_func(self):
        return self._root_par.val


# class LinkedParameter(ModelParameter):
#
#     def __init__(self,p_name,m_list):
#         super(LinkedParameter,self).__init__()
#         self.m_list = m_list
#         self.name = p_name
#
#     #def set_par(self, par_name, val):
#     #    for model in self.m_list:
#
#     #        pm = model.get_par_by_name(self.name)
#     #        pm.set(par_name, val = val)
#
#     #def set(self, *args, **kw):
#     #    self.set_par(*args, **kw)
#
# class ModelLinkedParameter(object):
#     def __init__(self,name,p):
#
#         self.name = name
#         self.parameters = ModelParameterArray()
#         self.paramters.add_par(p)
#         self.paramters.model=self





def compositr_parameter_setter(method):
    @wraps(method)
    def func_wrapper(self, model_name, *args, **kwargs):
        print('--> model_name',args,kwargs)
        try:
            if isinstance(model_name,str):
                pass
            else:
                model_name=model_name.name
            print('--> model_name', model_name, args,kwargs)
            return method(self, *args, **kwargs)
        except Exception as e:
           message = str(e)
           message += '\n'
           message += 'Starting from veriosn 1.2.0, FitModel is a CompositeModel, hence to set parameters ' \
                      'you have to pass as first paramter the model name or model object of the corresponing parameter e.g. \n' \
                      '''   
                            fit_model.set_par('model-name',value) 
                            OR 
                            fit_model.set_par(jet,value) 
                      '''




           raise RuntimeError(message)

    return func_wrapper

def create_a_function( **kwargs):

    def function_template(**kwargs):
        return

    return function_template


class CompositeModelParameterArray(object):

    def __repr__(self):
        return str(self.show_pars())

    def __init__(self,):

        self.all_frozen = False
        self._parameters = []
        self._model_comp = []

    def reset_dependencies(self):
        for p in self.par_array:
            p.reset_dependencies()




    def link_par(self,par_name,model_name_list,root_model_name):
        m_root=self.get_model_by_name(root_model_name)
        p_root=m_root.get_par_by_name(par_name)

        for m_name in model_name_list:
            dep_model=self.get_model_by_name(m_name)

            if dep_model is not None:
                dep_par = dep_model.get_par_by_name(par_name)
                if dep_par is not None:
                    print('==> par:',dep_par.name, 'from model:',  dep_model.name, 'linked to same parameter in model', m_root.name )
                    if p_root == dep_par:
                        raise RuntimeError(" root and linked parameter can't be the same")
                    if dep_par.immutable is True:
                        raise RuntimeError(" this parameter is already linked or dependent ")
                    if m_root==dep_par.model:
                        raise RuntimeError(" linked and root model must be different")
                    #p_root._linked_models.append(dep_model)
                    #exec(dep_par.name+'= dep_par')
                    #identity_func=lambda p_root=p_root: p_root.val_lin
                    dep_par._root_par = p_root
                    #dep_par.make_dependent_par(dep_par.identity_func)
                    dep_par._func=dep_par.identity_func
                    dep_par._linked = True
                    dep_par._is_dependent = True
                    dep_par._linked_root_model=m_root
                    dep_par._add_master_par(p_root)
                    p_root._add_depending_par(dep_par)
                    dep_par.freeze()
                    p_root.val=p_root.val
            else:
                 self._handle_missing_component_error(m_name)



    def _handle_missing_component_error(self, model_name):

        if hasattr(model_name, 'name'):
            name = model_name.name
        elif type(model_name) == str:
            name = model_name
        else:
            name = 'passed'
        s = 'Model component'+ name +' not present'
        raise RuntimeError(s)

    def add_model_parameters(self, model):
        try:
            assert (model.name not in [m.name for m in self._model_comp])
        except:
            raise RuntimeError('model name:', model.name, 'already assigned')


        model.parameters.model=model
        for p in model.parameters.par_array:
            p.model=model

        self._parameters.append(model.parameters)
        self._model_comp.append(model)

    def del_model_parameters(self, model_name):
        m,ID=self.get_model_by_name(model_name,get_idx=True)
        if m is not None:
            #_p = self._comp_par_array.pop(ID)
            _p = self._model_comp.pop(ID)
            _p = self._parameters.pop(ID)
        else:
             self._handle_missing_component_error(model_name)


    def get_model_by_name(self, model_name,get_idx=False):
        try:
            if isinstance(model_name, str):
                pass
            else:
                model_name = model_name.name

            selected_model = None
            idx=None
            for ID,m in enumerate(self._model_comp):
                if m.name == model_name:
                    selected_model = m
                    idx=ID


            if selected_model is None:
                self._handle_missing_component_error(model_name)
            if get_idx is False:
                return selected_model
            else:
                return selected_model,idx
        except Exception as e:
           message = str(e)
           message += '\n'
           message += 'Starting from veriosn 1.2.0, FitModel is a CompositeModel, hence to set parameters ' \
                      'you have to pass as first paramter the model name or model object of the corresponing parameter e.g. \n' \
                      '''   
                            fit_model.set_par('model-name',value) 
                            OR 
                            fit_model.set_par(jet,value) 
                      '''
    def get_par_by_name(self, model_name,par_name):
        p=None
        m=self.get_model_by_name(model_name)

        if m is None:
            self._handle_missing_component_error(model_name)
        else:
            p=m.parameters.get_par_by_name(par_name)

        return p


    def _build_par_table(self):
        _l=[]
        for ID, mc in enumerate(self._model_comp):
            self._parameters[ID]._build_par_table()
            if len(self._parameters[ID]._par_table) >= 1:
                #t=copy.copy(self._parameters[ID]._par_table)
                #for c in t.columns:
                 #   t[c] = t[c].astype(np.object)
                _l.append(self._parameters[ID]._par_table)
        #print('--> _l',_l)
        self._par_table=vstack(_l)

    def _build_best_fit_par_table(self):
        _l = []
        for ID, mc in enumerate(self._model_comp):
            self._parameters[ID]._build_best_fit_par_table()
            if len(self._parameters[ID]._best_fit_par_table) >= 1:
                #t = copy.copy(self._parameters[ID]._best_fit_par_table)
                #for c in t.columns:
                #    t[c] = t[c].astype(np.object)
                _l.append(self._parameters[ID]._best_fit_par_table)
        self._best_fit_par_table = vstack(_l)

    @property
    def par_array(self):
        pa=[]
        for p in  self._parameters:
            pa.extend(p.par_array)

        return pa

    @property
    def par_table(self):
        self._build_par_table()
        return self._par_table

    @property
    def best_fit_par_table(self):
        self._build_best_fit_par_table()
        return self._best_fit_par_table

    def show_pars(self, getstring=False, sort_key=None):
        self._build_par_table()
        if sort_key is not None:
            self.par_table.sort(sort_key)

        if getstring == True:
            return self.par_table.pformat_all()
        else:
            #self.par_table.pprint_all()
            _show_table(self.par_table)

    def show_best_fit_pars(self, getstring=False):
        self._build_best_fit_par_table()
        if getstring == True:
            return self._best_fit_par_table.pformat_all()
        else:
            #return self._best_fit_par_table
            #.pprint_all()
            _show_table(self._best_fit_par_table)

    #@compositr_parameter_setter
    def freeze(self, model_name,par_name):
        self.set(model_name,par_name, 'frozen')

    #@compositr_parameter_setter
    def free(self,model_name, par_name):
        self.set(model_name,par_name, 'free')

    #@compositr_parameter_setter
    def set(self, model_name, par_name, *args, **kw):
        m=self.get_model_by_name(model_name)
        if m is not None:
            m.parameters.set(par_name, *args, **kw)
        else:
            self._handle_missing_component_error(model_name)

    #@compositr_parameter_setter
    def set_par(self,model_name, par_name, val):
        m=self.get_model_by_name(model_name)
        if m is not None:
            m.parameters.set(par_name, val=val)
        else:
            self._handle_missing_component_error(model_name)

    #@compositr_parameter_setter
    def get(self, model_name,par_name, field_name, *args, **kw):
        #print('-->', par_name, field_name)
        m = self.get_model_by_name(model_name)
        if m is not None:
            pass
        else:
            self._handle_missing_component_error(model_name)
        #print('-->',par_name,field_name)
        return m.parameters.get(par_name, field_name, *args, **kw )

    #@compositr_parameter_setter
    def get_val(self, model_name,par_name):
        m = self.get_model_by_name(model_name)
        if m is not None:
            pass
        else:
            self._handle_missing_component_error(model_name)

        return m.parameters.get(par_name, 'val')

    def freeze_all(self):
        self.all_frozen = True
        for p_arr in self._parameters:
            for pi in range(len(p_arr)):
                self.par_array[pi].freeze()

    def free_all(self):
        self.all_frozen = False
        for p_arr in self._parameters:
            for pi in range(len(p_arr)):
                self.par_array[pi].free()


class ModelParameterArray(object):
    """
    This class provide and interface to handle an array of :class:`ModelParameter` objects.
    
    
    Attributes
    ----------
    par_array :  list 
        list of :class:`ModelParameter` objects
    """   

    def __repr__(self):
        return str(self.show_pars())

    #def __str__(self):
    #    return str(self.show_pars())

    def __init__(self,model=None):
            
        """
        Constructor 
        """
        
        self.par_array=[]
        self.all_frozen=False

        self.properties={}
        self.model=model
        self._numeric_fields = ['val', 'phys. bound. min', 'phys. bound. max','bestfit val','err +','err -','start val','fit range min','fit range max']


    def reset_dependencies(self):
        for p in self.par_array:
            p.reset_dependencies()

    def add_par(self,par):
        """
        adds a new :class:`ModelParameter` object  to the `par_array`
        """
        try:
            assert (isinstance(par,ModelParameter))
        except:
            raise RuntimeError('parameter is not an instance of',type(ModelParameter))

        try:
            assert (par.name not in [p.name for p in self.par_array])
        except:
            raise RuntimeError('parameter name:',par.name,'already assigned')

        par.model = self.model
        self.par_array.append(par)

        setattr(self,clean_var_name(par.name), par)
        self.properties[par.name]=par
        par.model=self.model
    @property
    def names(self):
        return [p.name for p in self.par_array]

    def _build_par_table(self,names_list=None):
        #, skip_hidden = False):
        _model_name = []
        _name=[]
        _type=[]
        _unit=[]
        _val=[]
        _bound_min=[]
        _bound_max=[]
        _islog=[]
        _frozen=[]
        _fields=[_model_name,_name,_type,_unit,_val,_bound_min,_bound_max,_islog,_frozen]
        _names=['model name','name','par type','units','val','phys. bound. min','phys. bound. max','log','frozen']


        if self.model is  None:
           _fields.pop(0)
           _names.pop(0)

        for par in self.par_array:

            append=False

            if names_list is not None:
                if par.name in names_list:
                    append=True
            else:
                append = True

            if type(par.val) is str:
                try:
                    ast.literal_eval(par.val)
                except:
                    append= False


            if append is True:
                if self.model is not None:
                    _model_name.append(self.model.name)
                else:
                    _model_name.append('no_name')

                if par._linked is True and par._linked_root_model != self.model:
                    _p_name = par.name+'(L,%s)'%par._linked_root_model.name

                    _type.append(par.par_type)
                    _unit.append(par.units)
                    _val.append(None)
                    _bound_min.append(None)
                    _bound_max.append(None)
                    _islog.append(par.islog)
                    _frozen.append(par.frozen)
                else:
                    if par._linked is False and par._depending_pars!=[]:
                        _p_name = par.name + '(M)'
                    else:
                        _p_name = par.name

                    _type.append(par.par_type)
                    _unit.append(par.units)
                    _val.append(par.val)
                    _bound_min.append(par.val_min)
                    _bound_max.append(par.val_max)
                    _islog.append(par.islog)
                    _frozen.append(par.frozen)

                if par._is_dependent is True and par._linked is False:
                    for p in par._master_pars:
                        _p_name = '*'+ par.name + '(D,%s)' % p.name
                    if par._linked_root_model is not None:
                        if par._linked_root_model != self.model:
                            _p_name +=  par._linked_root_model.name

                _name.append(_p_name)

        #_val = np.array(_val, dtype=np.object)
        t=Table(_fields,names=_names,masked=False)

        self._fromat_column_entry(t)

        self._par_table= t


    def _fromat_column_entry(self, t):
        for n in self._numeric_fields:
            if n in t.colnames:

                try:
                    if None in t[n].data:

                        t[n] = MaskedColumn(t[n].data, name=n, dtype=np.float, mask=t[n].data==None)
                    else:
                        t[n] = MaskedColumn(t[n].data, name=n, dtype=np.float)
                    t[n].format = '%e'
                except:

                    for ID,v in enumerate(t[n].data):
                        try:
                            c=ast.literal_eval(t[n].data[ID])
                            if type(c) == int:
                                t[n].data[ID] = '%d' % c
                            else:
                                t[n].data[ID] = '%e' % c
                        except:
                           pass




    def _build_best_fit_par_table(self, names_list=None):
        # skip_hidden=False):

        _name=[]
        _model_name = []
        _best_fit_val=[]
        _best_fit_err_p=[]
        _best_fit_err_m=[]
        _val_start=[]
        _fit_range_min=[]
        _fit_range_max=[]
        _frozen=[]
        _val=[]
        _fields=[_model_name,_name,_val,_best_fit_val,_best_fit_err_p,_best_fit_err_m,_val_start,_fit_range_min,_fit_range_max,_frozen]
        _names=['model name','name','val','bestfit val','err +','err -','start val','fit range min','fit range max','frozen']

        if self.model is None:
           _fields.pop(0)
           _names.pop(0)

        for par in self.par_array:

            append=False

            if names_list is not None:
                if par.name in names_list:
                    append=True
            else:
                append = True

            if type(par.val_start) is str:
                try:
                    ast.literal_eval(par.val_start)
                except:
                    append= False

            if append:
                if self.model is not None:
                    _model_name.append(self.model.name)
                else:
                    _model_name.append('no_name')

                if par._linked is True and par._linked_root_model != self.model:
                    _p_name = par.name + '(L,%s)' % par._linked_root_model.name

                    _val_start.append(None)
                    _best_fit_val.append(None)
                    _best_fit_err_p.append(None)
                    _best_fit_err_m.append(None)

                    _fit_range_min.append(par.fit_range_min)
                    _fit_range_max.append(par.fit_range_max)

                else:
                    if par._linked is False and par._depending_pars!=[]:
                        _p_name = par.name + '(M)'
                    else:
                        _p_name = par.name
                    if par.frozen == True:
                        best_fit_val = None
                        best_fit_err_p = None
                        best_fit_err_m = None
                    else:
                        best_fit_val=par.best_fit_val
                        if hasattr(par, 'err_p') and hasattr(par, 'err_m'):
                            best_fit_err_m =  par.err_m
                            best_fit_err_p =  par.err_p
                        else:
                            best_fit_err_m = None
                            if par.best_fit_err is None:
                                best_fit_err_p = None
                            else:
                                best_fit_err_p =  par.best_fit_err

                    _val_start.append(par.val_start)
                    _best_fit_val.append(best_fit_val)
                    _best_fit_err_p.append(best_fit_err_p)
                    _best_fit_err_m.append(best_fit_err_m)

                    _fit_range_min.append(par.fit_range_min)
                    _fit_range_max.append(par.fit_range_max)

                if par._is_dependent is True and par._linked is False:
                    for p in par._master_pars:
                        _p_name = '*' + par.name + '(D,%s)' % p.name
                    if par._linked_root_model is not None:
                        if par._linked_root_model != self.model:
                            _p_name += par._linked_root_model.name

                _name.append(_p_name)
                _frozen.append(par.frozen)
                _val.append(par.val)

        #_val_start = np.array(_val_start, dtype=np.object)
        #_best_fit_val = np.array(_val_start, dtype=np.object)

        # set false to avoid string in hidden
        t = Table(_fields, names=_names, masked=False)

        self._fromat_column_entry(t)

        self._best_fit_par_table= t



    def __setattr__(self, name, value):
        if "properties" in self.__dict__ and name in self.properties:
            raise AttributeError('this member is protected, use del_par,add_par, to add/remove, and set() to set values or .val attribute')
        else:
            self.__dict__[name] = value


    def del_par(self,par):
        
        self.par_array.remove(par)
        delattr(self,par.name)
        self.properties.pop(par.name)

    def get_par_by_name(self,name, verbose=False):
        """
        selects a parameter by name
       
        Parameters
        ----------
        name : (str) parameter name
        
        Returns
        -------
        item : the :class:`ModelParameter` element of the `par_array` with the corresponding name
        
        """
        for pi in range(len(self.par_array)):
            if self.par_array[pi].name==name:
                return self.par_array[pi]
        else:
            if verbose:
                print ("no par with name %s found"%name)
                print ("pars in array are:")
                self.show_pars()
            
            return None
                
    def get_par_by_type(self,par_type):
        """

        get parameter by type

        """
        for param in self.par_array:
            if param.par_type==par_type:
                return param

        return None

    @property
    def par_table(self):
        self._build_par_table()
        return self._par_table

    @property
    def best_fit_par_table(self):
        self._build_best_fit_par_table()
        return self._best_fit_par_table

    def show_pars(self,getstring=False,names_list=None,sort_key=None):

        self._build_par_table(names_list=names_list)
        if sort_key is not None:
            self.par_table.sort(sort_key)

        if getstring==True:
            return self.par_table.pformat_all()
        else:
            _show_table(self.par_table)
            #return self.par_table
            #.pprint_all()
        
    
    
    def show_best_fit_pars(self,getstring=False):

        self._build_best_fit_par_table()
        if getstring == True:
            return self.best_fit_par_table.pformat_all()
        else:
            return self.best_fit_par_table
            #.pprint_all()


    def set(self,par_name,*args, **keywords):
        
        """
        sets to a given value a given parameter
        
        Parameters
        ----------
        
        par_name     : (str) name of the parameter 
        keywords     :  keywords to set the value or the range of the parameter 
            
         
         
           
        Examples
        --------
        if parameters is a :class:`ModelParameterArray` object:
        
        .. code::
     
            parameters.set('R',val=1E16) 
            parameters.set('R',fit_range=[1E16,1E17])
        
        

        """
        
        #print "ModelParamterArray in model parameters",args,keywords
        
        par=self.get_par_by_name(par_name)

        args_list=['free','frozen']

        if  args!=():
            for  arg in args:
            
                if arg not in args_list:
                    print ("argument: %s,  not in allowed args"%arg)
                    print ("allowed args= ",args_list)
                    return
                
                 
                if arg=='frozen':                
                    par.freeze()
                
                
                if  arg=='free':
                    par.free()

                
            
        
        if keywords!={}:
        
            if par is None:
                raise RuntimeWarning('parameter %s is not present in the model'%par_name)
            else:
                par.set( **keywords)
        
        
        
    def get(self,par_name,arg):
        
        """
        gets the argument of a given parameter
        
        Parameters
        ----------
        
        par_name     : (str) name of the parameter 
        arg     :  keyword
            
         
         
           
        Examples
        --------
        if parameters is a :class:`ModelParameterArray` object:
        
        .. code::
     
            parameters.get('R') 
            parameters.get('frozen')
        
        

        """
        
        #print "ModelParamterArray in model parameters",args,keywords
        
        par=self.get_par_by_name(par_name)
        
        return par.get(arg)
    
    def freeze_all(self):
        self.all_frozen=True
        for pi in range(len(self.par_array)):
            self.par_array[pi].freeze()
                
        
    def free_all(self):
        self.all_frozen = False
        for pi in range(len(self.par_array)):
            self.par_array[pi].free()
        

    def _serialize_pars(self):
        _par_keys=['val','val_min','val_max','val_start','val_last_call','fit_range_min','fit_range_max','best_fit_val',
                   'best_fit_err','frozen','allowed_values','_linked','_is_dependent','_func','_master_pars',
                   '_linked_root_model','_depending_pars','_root_par','','_master_par_list','_depending_par_expr','units','par_type']
        _par_dict = {}
        for par in self.par_array:
            _val_dict={}

            for k in _par_keys:
                if hasattr(par,k):
                    _val_dict[k]=getattr(par,k)

            _par_dict[par.name] = _val_dict

        return _par_dict

    def _decode_pars(self,_par_dict):
        for p_name in _par_dict.keys():
            self.set(p_name,**_par_dict[p_name])




__author__ = "Andrea Tramacere"

import os
import numpy as np
from inspect import signature
from astropy.constants import m_e,m_p,c
from scipy import interpolate
from .jetkernel_models_dic import gamma_dic_e ,gamma_dic_p, gamma_dic_pp_e_second, available_N_distr, N_distr_descr, available_emitters_type
from .plot_sedfit import PlotPdistr
from .jet_paramters import *
from .utils import set_str_attr
from .model_parameters import ModelParameterArray, ModelParameter

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
if on_rtd is True:
    try:
        from .jetkernel import jetkernel as BlazarSED
    except ImportError:
        from .mock import jetkernel as BlazarSED
else:
    from .jetkernel import jetkernel as BlazarSED

__all__=['EmittersDistribution','BaseEmittersDistribution', 'ArrayDistribution','EmittersArrayDistribution','InjEmittersDistribution','JetkernelEmittersDistribution']


class ArrayDistribution(object):

    def __init__(self, e_array, n_array, gamma_grid_size=None):
        self.e_array = e_array
        self.n_array = n_array
        _size = e_array.size

        if n_array.size != _size:
            raise RuntimeError('e_array and n_array must have same size')
        self.size = _size
        if gamma_grid_size is None:
                self.gamma_grid_size = _size * 2 + 1


def check_par_name(method):
    def inner(ref,  *args, **kwargs):
        #print(ref,args,kwargs)
        if args[0]  in ref._skip:
            raise RuntimeError('par name',args[0],'can not be in proteced names',ref._skip)
        else:
            return method(ref,*args, **kwargs)

    return inner


class BaseEmittersDistribution(object):

    def __init__(self,
                 name,
                 spectral_type,
                 gamma_grid_size=200,
                 log_values=False,
                 emitters_type='electrons',
                 skip_build=False,
                 normalize=False):

        self._spectral_type = None
        self._allowed_spectral_types = ['bkn', 'plc', 'lp', 'lppl', 'pl', 'lpep', 'array','user_defined']
        self._set_emitters_type(emitters_type)
        self._set_spectral_type(spectral_type)
        self._Norm = 1.0
        self._skip = ['gmin', 'gmax', 'N', 'NH_pp', 'Q']
        if skip_build is False:
            self._build( name, log_values, gamma_grid_size, normalize)

    def _build(self, name, log_values, gamma_grid_size, normalize):
        self._user_defined = True
        self._name = name
        self._log_values = log_values
        self._gamma_grid_size = gamma_grid_size
        self.parameters = ModelParameterArray()
        self._set_base_pars()
        self.normalize = normalize
        self.gamma_cooling_eq_second = None
        self._secondaries_done = False

    def _set_spectral_type(self,spectral_type):
        if spectral_type not in self._allowed_spectral_types:
            raise RuntimeError('spectral_type=',spectral_type,' not in allowed',self._allowed_spectral_types)
        else:
            self._spectral_type = spectral_type


    @staticmethod
    def spectral_types_obs_constrain():
        return ['bkn', 'plc', 'lp', 'lppl', 'pl', 'lpep', 'array']

    @property
    def spectral_type(self):
        return self._spectral_type

    @check_par_name
    def add_par(self, name, par_type, val, vmax, vmin, unit='', log=False, frozen=False):
        #print('==> adding',name,val)
        #if log is True:
        #    val = np.log10(val)

        self.parameters.add_par(ModelParameter(name=name,
                                               par_type=par_type,
                                               val=val,
                                               val_min=vmin,
                                               val_max=vmax,
                                               units=unit,
                                               frozen=frozen,
                                               log=log,))

    def set_distr_func(self,distr_func):
        if distr_func is not None:
            self._validate_func(distr_func)
        self.distr_func = distr_func

    def _validate_func(self,distr_func):
        s=signature(distr_func)
        for param in s.parameters.values():
            if self.parameters.get_par_by_name(param.name) is None and param.name!='gamma':
                raise RuntimeError('distr_func parameter',param.name,'not among parameters names',self.parameters.names)
        if 'gamma' not in s.parameters.keys():
            raise RuntimeError('gamma not present in distr_func signature')

    def _set_emitters_type(self,emitters_type):
        self.available_emitters_type = available_emitters_type
        self._check_emitters_type(emitters_type)
        self.emitters_type = emitters_type


    def _set_base_pars(self):
        #model_dic = {}
        a_h, b_h = self.set_bounds(1, 1E15, log_val=self._log_values)
        a_l, b_l = self.set_bounds(1, 1E9, log_val=self._log_values)
        #a_t, b_t = self.set_bounds(1, 1E9, log_val=self._log_values)

        self.parameters.add_par( ModelParameter(name='gmin', par_type='low-energy-cut-off',val=2, val_min=a_l, val_max=b_l,
                                                  units='lorentz-factor', log=self._log_values))

        self.parameters.add_par( ModelParameter(name='gmax', par_type='high-energy-cut-off',val=1E5, val_min=a_h, val_max=b_h,
                                                  units='lorentz-factor', log=self._log_values))

        self.parameters.add_par( ModelParameter(name='N', par_type='emitters_density', val=1, val_min=0, val_max=None, units='cm^-3'))


    def set_bounds(self,a,b,log_val=False):
        if log_val == False:
            return [a,b]

        else:
            return np.log10([a,b])

    def _set_log_val(self,a,log_val=False):
        if log_val == False:
            return a

        else:
            return np.log10(a)


    def _eval_func(self,gamma):
        p_dict = {}
        for par in self.parameters.par_array:
            if par.name not in self._skip:
                p_dict[par.name] = par.val_lin
        p_dict['gamma']=gamma

        return self.distr_func(**p_dict)

    @property
    def name(self):
        return self._name

    def _check_emitters_type(self, emitters_type):
        if emitters_type not in available_emitters_type:
            raise RuntimeError("type of emitters :%s, not allowed" % emitters_type, "please choose among: ",
                               self.available_emitters_type)
        else:
            pass

    def update(self):
        self._fill()

    def set_grid(self):
        gmin = self.parameters.get_par_by_name('gmin').val_lin
        gmax = self.parameters.get_par_by_name('gmax').val_lin
        self._gamma_grid = np.logspace(np.log10(gmin), np.log10(gmax), self._gamma_grid_size)

    def eval_N(self):
        self.update()
        if self.emitters_type=='electrons':
            return np.trapz(self.n_gamma_e,self.gamma_e)
        elif self.emitters_type=='protons':
            return np.trapz(self.n_gamma_p, self.gamma_p)
        else:
            raise  RuntimeError('emitters type',self.emitters_type, 'not valid')

    def eval_U(self, gmin=None, gmax=None):
        self.update()
        if self.emitters_type == 'electrons':
            x=self.gamma_e
            y=self.n_gamma_e * self.gamma_e
            cost=m_e.cgs.value * (c.cgs ** 2).value

        elif self.emitters_type == 'protons':
            x = self.gamma_p
            y = self.n_gamma_p * self.gamma_p
            cost =  m_p.cgs.value * (c.cgs ** 2).value

        else:
            raise  RuntimeError('emitters type',self.emitters_type, 'not valid')

        msk = np.ones(x.shape, dtype=np.bool)
        if gmin is not None:
            msk = x>=gmin
        if gmax is not None:
            msk = np.logical_and(msk, x<=gmax)

        return np.trapz(y[msk],x[msk]) * cost

    def _fill(self,):
        self.set_grid()

        self.f=self._eval_func(gamma=self._gamma_grid)
        self._Norm=1.0

        if self.normalize is True:
            self._Norm=1.0/np.trapz(self.f,self._gamma_grid)

        self.f = self.f*self._Norm*self.parameters.get_par_by_name('N').val

        if self.emitters_type == 'electrons':
            self.gamma_e = np.zeros(self._gamma_grid_size)
            self.n_gamma_e = np.zeros(self._gamma_grid_size)
            self.gamma_e = self._gamma_grid
            self.n_gamma_e = self.f

            self.n_gamma_e[np.isnan(self.n_gamma_e)] = 0
            self.n_gamma_e[np.isinf(self.n_gamma_e)] = 0

        if self.emitters_type == 'protons':
            self.gamma_p = np.zeros(self._gamma_grid_size)
            self.n_gamma_p = np.zeros(self._gamma_grid_size)
            if self._secondaries_done is False:
                self.gamma_e_second_inj = np.zeros(self._gamma_grid_size)
                self.n_gamma_e_second_inj = np.zeros(self._gamma_grid_size)

            self.gamma_p = self._gamma_grid
            self.n_gamma_p = self.f

            self.n_gamma_p[np.isnan(self.n_gamma_p)] = 0
            self.n_gamma_p[np.isinf(self.n_gamma_p)] = 0

            self.n_gamma_e_second_inj[np.isnan(self.n_gamma_e_second_inj)] = 0
            self.n_gamma_e_second_inj[np.isinf(self.n_gamma_e_second_inj)] = 0


    def _plot(self,m, p, y_min=None, y_max=None, x_min=None, x_max=None, energy_unit='gamma',label=None):

        if hasattr(self,'_jet'):
            if self._jet is not None:
                self._set_blob()
        self.update()

        if self.emitters_type == 'electrons':
            if label is None:
                label = 'electrons'
            m(self.gamma_e,
                         self.n_gamma_e,
                         y_min=y_min,
                         y_max=y_max,
                         x_min=x_min,
                         x_max=x_max,
                         particle='electrons',
                         energy_unit=energy_unit,
                         label=label)
        if self.emitters_type == 'protons':
            if label is None:
                label = 'protons'
            m(self.gamma_p,
                         self.n_gamma_p,
                         y_min=y_min,
                         y_max=y_max,
                         x_min=x_min,
                         x_max=x_max,
                         particle='protons',
                         energy_unit=energy_unit,
                         label=label)
            if self._secondaries_done is True:
                if hasattr(self,'n_gamma_e'):
                    m(self.gamma_e,
                                 self.n_gamma_e,
                                 y_min=y_min,
                                 y_max=y_max,
                                 x_min=x_min,
                                 x_max=x_max,
                                 particle='electrons',
                                 energy_unit=energy_unit,
                                 label='electrons sec.')
                    if energy_unit != 'gamma':

                        eq= self.gamma_cooling_eq_second* (m_e * c * c).to(energy_unit).value
                    else:
                        eq = self.gamma_cooling_eq_second

                    p.ax.axvline(np.log10(eq), ls='--', label='cooling. eq. second.', lw=0.5, c='r')

                if hasattr(self, 'gamma_e_second_inj'):
                    m(self.gamma_e_second_inj,
                      self.n_gamma_e_second_inj,
                      y_min=y_min,
                      y_max=y_max,
                      x_min=x_min,
                      x_max=x_max,
                      particle='electrons',
                      energy_unit=energy_unit,
                      label='electrons sec. (inj)')

        return p

    def plot(self, p=None, y_min=None, y_max=None, x_min=None, x_max=None, energy_unit='gamma',label=None,loglog=False):
        if p is None:
            p = PlotPdistr(loglog=loglog)
        m=getattr(p,'plot_distr')
        self._plot(m,p,y_min=y_min,y_max=y_max,x_min=x_min,x_max=x_max,energy_unit=energy_unit,label=label)
        return p

    def plot2p(self, p=None, y_min=None, y_max=None, x_min=None, x_max=None, energy_unit='gamma',label=None,loglog=False):
        if p is None:
            p = PlotPdistr(loglog=loglog)
        m=getattr(p,'plot_distr2p')
        self._plot(m,p,y_min=y_min,y_max=y_max,x_min=x_min,x_max=x_max,energy_unit=energy_unit,label=label)

        return p

    def plot3p(self, p=None, y_min=None, y_max=None, x_min=None, x_max=None, energy_unit='gamma',label=None,loglog=False):
        if p is None:
            p = PlotPdistr(loglog=loglog)
        m=getattr(p,'plot_distr3p')
        self._plot(m,p,y_min=y_min,y_max=y_max,x_min=x_min,x_max=x_max,energy_unit=energy_unit,label=label)
        return p


class EmittersDistribution(BaseEmittersDistribution):

    def __init__(self,
                 name,
                 spectral_type,
                 jet=None,
                 gamma_grid_size=200,
                 log_values=False,
                 emitters_type='electrons',
                 normalize=False,
                 skip_build=False):

        super(EmittersDistribution, self).__init__(name,
                                                   spectral_type=spectral_type,
                                                   emitters_type=emitters_type,
                                                   gamma_grid_size=gamma_grid_size,
                                                   skip_build=True,
                                                   normalize=normalize,
                                                   log_values=log_values)


        if isinstance(self,EmittersArrayDistribution):
            self._array_gamma = None
            self._array_n_gamma = None
        else:
            if skip_build is False:
                self._build(jet,name,log_values, gamma_grid_size,normalize)

    def _copy_from_jet(self, jet):
        self._name = jet.emitters_distribution._name
        self._log_values = jet.emitters_distribution._log_values
        self._gamma_grid_size = jet.emitters_distribution._log_values
        self._gamma_grid_size = jet.emitters_distribution._gamma_grid_size
        self.normalize = jet.emitters_distribution.normalize

        for par in self.parameters.par_array:
            p = jet.emitters_distribution.parameters.get_par_by_name(par.name)
            par.set(val=p.val, skip_dep_par_warning=True)

    def _build(self,jet,name,log_values, gamma_grid_size,normalize):
        self._user_defined=True
        self._name = name
        self._log_values = log_values
        self._gamma_grid_size=gamma_grid_size
        self.parameters = ModelParameterArray()
        self.set_parameters_dict()
        self.normalize = normalize
        self.gamma_cooling_eq_second=None
        self._secondaries_done=False
        self.set_jet(jet)


    def _update_parameters_dict(self):
        for par in self.parameters.par_array:
            self._parameters_dict[par.name].val = par.val
            self._parameters_dict[par.name].vmin = par.val_min
            self._parameters_dict[par.name].vmax = par.val_max
            self._parameters_dict[par.name].log = par.islog
            self._parameters_dict[par.name].punit = par.units
            self._parameters_dict[par.name]._is_dependent = par._is_dependent
            self._parameters_dict[par.name]._depending_pars = par._depending_pars
            self._parameters_dict[par.name]._func = par._func
            self._parameters_dict[par.name]._master_pars = par._master_pars

    def add_par(self, name, par_type, val, vmax, vmin, unit='', log=False, frozen=False):
        #if log is True:
        #    val = np.log10(val)

        if name not in self._parameters_dict.keys():
            self._parameters_dict[name]=JetModelDictionaryPar(ptype=par_type,
                                                              vmin=vmin,
                                                              vmax=vmax,
                                                              log=log,
                                                              val=val,
                                                              punit=unit,
                                                              froz=frozen,
                                                              is_in_jetkernel=False)
        else:
            raise ValueError('par',name,'already assigned')

        #print('==> add par',name,val,log)
        self.parameters.add_par(ModelParameter(name=name,
                                               par_type=par_type,
                                               val=val,
                                               val_min=vmin,
                                               val_max=vmax,
                                               units=unit,
                                               frozen=frozen,
                                               log=log,))

    def set_parameters_dict(self,):
        model_dict, a_h, b_h, a_l, b_l, a_t, b_t = self._get_base_dict_and_bounds()
        model_dict['gmin'].val = 2.0
        model_dict['gmin'].is_in_jetkernel = True

        model_dict['gmax'].val = 1E6
        model_dict['gmax'].is_in_jetkernel = True
        if isinstance(self,EmittersArrayDistribution):
            model_dict['N'].val = 1.0
        else:
            model_dict['N'].val = 100.
        model_dict['N'].is_in_jetkernel = True

        if 'NH_pp' in model_dict.keys():
            model_dict['NH_pp'].val = 1.
            model_dict['NH_pp'].is_in_jetkernel = True

        self._parameters_dict=model_dict

        for k,par in model_dict.items():
            # print('==> k',k,'par',par,par.val,par.log)
            if par.log is True:
                val = np.log10(par.val)
            else:
                val = par.val
            # print('==> k', k, 'par', par, val, par.log)
            self.parameters.add_par(ModelParameter(name=k,
                                                   par_type=par.ptype,
                                                   val=val,
                                                   val_min=par.vmin,
                                                   val_max=par.vmax,
                                                   units=par.punit,
                                                   frozen=par.froz,
                                                   log=par.log))

    def _set_base_pars(self):
        raise RuntimeError('allowed only in parent class')

    def _get_base_dict_and_bounds(self):
        model_dic = {}
        a_h, b_h = self.set_bounds(1, 1E15, log_val=self._log_values)
        a_l, b_l = self.set_bounds(1, 1E9, log_val=self._log_values)
        a_t, b_t = self.set_bounds(1, 1E9, log_val=self._log_values)
        model_dic['gmin'] = JetModelDictionaryPar(ptype='low-energy-cut-off', vmin=a_l, vmax=b_l,
                                                  punit='lorentz-factor', log=self._log_values)
        model_dic['gmax'] = JetModelDictionaryPar(ptype='high-energy-cut-off', vmin=a_h, vmax=b_h,
                                                  punit='lorentz-factor', log=self._log_values)

        model_dic['N'] = JetModelDictionaryPar(ptype='emitters_density', vmin=0, vmax=None, punit='cm^-3')

        if self.emitters_type =='protons':
            model_dic['NH_pp'] = JetModelDictionaryPar(ptype='target_density', vmin=0, vmax=None, punit='cm^-3',froz =False)

        if isinstance(self, EmittersArrayDistribution):
            model_dic['N'] = JetModelDictionaryPar(ptype='scaling_factor', vmin=0, vmax=None, punit='')
        return model_dic, a_h, b_h, a_l, b_l, a_t, b_t

    def set_jet(self, jet):
        if jet is not None:

            #name passed to the C code
            #do not change
            name = 'jetset'

            self._jet = jet
            set_str_attr(jet._blob, 'DISTR', name)
            set_str_attr(jet._blob, 'PARTICLE', self.emitters_type)

            p = self._jet.get_par_by_name('gmin')
            if p is not None:
                p.set(value=self.parameters.get_par_by_name('gmin').val_lin)
            else:
                p=self.parameters.get_par_by_name('gmin')
            setattr(jet._blob, 'gmin', p.val_lin)

            p = self._jet.get_par_by_name('gmax')
            if p is not None:
                p.set(value=self.parameters.get_par_by_name('gmax').val_lin)
            else:
                p = self.parameters.get_par_by_name('gmax')
            setattr(jet._blob, 'gmax', p.val_lin)

            self._jet._blob.gamma_grid_size = self._gamma_grid_size
        else:
            self._jet=jet


    def set_grid_size(self,gamma_grid_size):
        if gamma_grid_size is not None:
            if self._jet is not  None:
              setattr(self._jet._blob,'gamma_grid_size' ,gamma_grid_size)
        self._fill()

    def set_grid(self):
        if self._jet is None:
            gmin = self.parameters.get_par_by_name('gmin').val_lin
            gmax = self.parameters.get_par_by_name('gmax').val_lin
            self._gamma_grid = np.logspace(np.log10(gmin), np.log10(gmax), self._gamma_grid_size)
        else:
            BlazarSED.setNgrid(self._jet._blob)
            if self.emitters_type == 'electrons':
                #print('==> Build Ne start')
                BlazarSED.build_Ne_jetset(self._jet._blob)
                gamma_ptr = getattr(self._jet._blob, 'griglia_gamma_jetset_Ne_log')
                #print('==> Build Ne done')
            elif self.emitters_type == 'protons':
                BlazarSED.build_Np_jetset(self._jet._blob)
                gamma_ptr = getattr(self._jet._blob, 'griglia_gamma_jetset_Np_log')
            else:
                raise RuntimeError('emitters type', self.emitters_type, 'not valid')

            size = self._jet._blob.gamma_grid_size
            self._gamma_grid = np.zeros(size)
            for ID in range(size):
                self._gamma_grid[ID] = BlazarSED.get_elec_array(gamma_ptr, self._jet._blob, ID)
            self._gamma_grid_size=self._jet.gamma_grid_size




    def _fill(self,):
        super(EmittersDistribution, self)._fill()
        if self._jet is not None:

            if self.emitters_type == 'electrons':
                size = self._gamma_grid_size
                Ne_ptr = getattr(self._jet._blob, 'Ne_jetset')

                for ID in range(size):
                    BlazarSED.set_elec_array(Ne_ptr, self._jet._blob, self.f[ID], ID)

            if self.emitters_type == 'protons':
                size = self._gamma_grid_size
                Np_ptr = getattr(self._jet._blob, 'Np_jetset')
                for ID in range(size):
                    BlazarSED.set_elec_array(Np_ptr, self._jet._blob, self.f[ID], ID)



    def _set_blob(self):
        if self.emitters_type == 'electrons':

            self._Ne_name, self._gammae_name = gamma_dic_e['electron_distr']
            self.Ne_ptr = getattr(self._jet._blob, self._Ne_name)
            self.e_gamma_ptr = getattr(self._jet._blob, self._gammae_name)

        elif self.emitters_type == 'protons':
            self._Ne_name, self._gammae_name = gamma_dic_e['electron_distr']
            self._Np_name, self._gammap_name = gamma_dic_p['proton_distr']
            self._Q_inj_e_second_name, self._gammae_inj_sec_name = gamma_dic_pp_e_second['e_second_inj']

            self.Np_ptr = getattr(self._jet._blob, self._Np_name)
            self.p_gamma_ptr = getattr(self._jet._blob, self._gammap_name)
            self.Ne_ptr = getattr(self._jet._blob, self._Ne_name)
            self.e_gamma_ptr = getattr(self._jet._blob, self._gammae_name)

            self._Q_inj_e_second_ptr = getattr(self._jet._blob, self._Q_inj_e_second_name)
            self.e_inj_second_gamma_ptr = getattr(self._jet._blob, self._gammae_inj_sec_name)

        size = self._jet._blob.gamma_grid_size
        self.gamma_e = np.zeros(size)
        self.n_gamma_e = np.zeros(size)
        self.gamma_p = np.zeros(size)
        self.n_gamma_p = np.zeros(size)
        self.gamma_e_second_inj = np.zeros(size)
        self.n_gamma_e_second_inj = np.zeros(size)
        self.gamma_e_second_inj = np.zeros(size)
        self.q_gamma_e_second_inj = np.zeros(size)

        for ID in range(size):
            self.gamma_e[ID] = BlazarSED.get_elec_array(self.e_gamma_ptr, self._jet._blob, ID)
            self.n_gamma_e[ID] = BlazarSED.get_elec_array(self.Ne_ptr, self._jet._blob, ID)

        if self.emitters_type == 'protons':
            for ID in range(size):
                self.gamma_p[ID] = BlazarSED.get_elec_array(self.p_gamma_ptr, self._jet._blob, ID)
                self.n_gamma_p[ID] = BlazarSED.get_elec_array(self.Np_ptr, self._jet._blob, ID)

            for ID in range(size):

                self.gamma_e_second_inj[ID] = BlazarSED.get_elec_array(self.e_inj_second_gamma_ptr, self._jet._blob, ID)
                self.n_gamma_e_second_inj[ID] = BlazarSED.get_elec_array(self._Q_inj_e_second_ptr, self._jet._blob, ID)
            self.gamma_cooling_eq_second= self._jet._blob.gamma_cooling_eq
            self._secondaries_done = True




class EmittersArrayDistribution(EmittersDistribution):
    def __init__(self,
                 name,
                 jet=None,
                 emitters_type='electrons',
                 normalize=False,
                 skip_build=False,
                 gamma_array=None,
                 n_gamma_array=None,
                 gamma_grid_size=None):

        super(EmittersArrayDistribution, self).__init__(name,
                                                        spectral_type='array',
                                                        emitters_type=emitters_type)

        if gamma_array is not None and n_gamma_array is not None:
            if self._spectral_type != 'array':
                raise RuntimeError('you can pass  gamma_array and n_gamma_array only for array distribution')
            self._set_arrays(gamma_array,n_gamma_array)
        elif (gamma_array is not None) and (n_gamma_array is not None) is False:
            raise RuntimeError('you have to pass both gamma_array and n_gamma_array for array distribution')

        if gamma_grid_size is None:
            gamma_grid_size=gamma_array.size
        if skip_build is False:
            self._build(jet, name, False, gamma_grid_size, normalize)
        self.parameters.gmin.val=self._array_gamma[0]
        self.parameters.gmax.val = self._array_gamma[-1]
        self.parameters.N.val=1
        self.set_distr_func(self._array_func)

    def _set_arrays(self,gamma_array,n_gamma_array):
        _ids = np.argsort(gamma_array)
        self._array_gamma = gamma_array[_ids]
        self._array_n_gamma = n_gamma_array[_ids]

    def _array_func(self,gamma):
        msk_nan=self._array_n_gamma>0
        if msk_nan.sum()>1:
            f_interp = interpolate.interp1d(np.log10(self._array_gamma[msk_nan]), np.log10(self._array_n_gamma[msk_nan]), bounds_error=False, kind='linear')
            y = np.power(10., f_interp(np.log10(gamma)))
            msk_nan = np.isnan(y)
            y[msk_nan] = 0
        else:
            y = np.zeros(gamma.size)

        return y


class InjEmittersDistribution(BaseEmittersDistribution):

    def __init__(self,
                 name,
                 spectral_type,
                 gamma_grid_size=200,
                 log_values=False,
                 skip_build=False,
                 normalize=True,
                 emitters_type='electrons'):

        super(InjEmittersDistribution, self).__init__(name=name,
                                                      spectral_type=spectral_type,
                                                      emitters_type='electrons',
                                                      gamma_grid_size=gamma_grid_size,
                                                      skip_build=True,
                                                      normalize=True,
                                                      log_values=log_values)

        if skip_build is False:
            self._build(name, log_values, gamma_grid_size, normalize=True)


    def _set_base_pars(self):
        #model_dic = {}
        a_h, b_h = self.set_bounds(1, 1E15, log_val=self._log_values)
        a_l, b_l = self.set_bounds(1, 1E9, log_val=self._log_values)
        #a_t, b_t = self.set_bounds(1, 1E9, log_val=self._log_values)

        self.parameters.add_par( ModelParameter(name='gmin', par_type='low-energy-cut-off',val=2, val_min=a_l, val_max=b_l,
                                                  units='lorentz-factor', log=self._log_values))

        self.parameters.add_par( ModelParameter(name='gmax', par_type='high-energy-cut-off',val=1E5, val_min=a_h, val_max=b_h,
                                                  units='lorentz-factor', log=self._log_values))

        self.parameters.add_par( ModelParameter(name='Q', par_type='emitters_density', val=1E-3, val_min=0, val_max=None, units='cm-3 s-1'))

    def _build(self, name, log_values, gamma_grid_size, normalize):
        self._user_defined = True
        self._name = name
        self._log_values = log_values
        self._gamma_grid_size = gamma_grid_size
        self.parameters = ModelParameterArray()
        self._set_base_pars()
        self.normalize = normalize


    def _fill(self,):
        self.set_grid()

        self.f=self._eval_func(gamma=self._gamma_grid)
        self._Norm=1.0

        if self.normalize is True:
            self._Norm=1.0/np.trapz(self.f,self._gamma_grid)

        self.f = self.f*self._Norm*self.parameters.get_par_by_name('Q').val

        if self.emitters_type == 'electrons':
            self.gamma_e = np.zeros(self._gamma_grid_size)
            self.n_gamma_e = np.zeros(self._gamma_grid_size)
            self.gamma_e = self._gamma_grid
            self.n_gamma_e = self.f

            self.n_gamma_e[np.isnan(self.n_gamma_e)] = 0
            self.n_gamma_e[np.isinf(self.n_gamma_e)] = 0

    def eval_U_q(self):
        return self.eval_U()

    def _set_L_inj(self, L_inj_target_erg, volume):
        if L_inj_target_erg >0:
            self.parameters.Q.val *=L_inj_target_erg/(self.eval_U_q() * volume )
        else:
            pass
        #print('==>',self.eval_U()*volume*delta_t)
        self.update()

    def set_temp_ev(self):
        self.e_gamma_ptr = getattr(self._temp_ev, self._gammae_name)
        self._Q_inj_e_second_ptr = getattr(self._temp_ev._blob, self._Q_inj_e_second_name)



class InjEmittersArrayDistribution(InjEmittersDistribution):
    def __init__(self,
                 name,
                 emitters_type='electrons',
                 normalize=False,
                 skip_build=False,
                 gamma_array=None,
                 n_gamma_array=None,
                 gamma_grid_size=None):

        super(InjEmittersArrayDistribution, self).__init__(name,
                                                        spectral_type='array',
                                                        emitters_type=emitters_type)

        if gamma_array is not None and n_gamma_array is not None:
            if self._spectral_type != 'array':
                raise RuntimeError('you can pass  gamma_array and n_gamma_array only for array distribution')
            self._set_arrays(gamma_array,n_gamma_array)
        elif (gamma_array is not None) and (n_gamma_array is not None) is False:
            raise RuntimeError('you have to pass both gamma_array and n_gamma_array for array distribution')

        if gamma_grid_size is None:
            gamma_grid_size=gamma_array.size
        if skip_build is False:
            self._build( name, False, gamma_grid_size, normalize)
        self.parameters.gmin.val=self._array_gamma[0]
        self.parameters.gmax.val = self._array_gamma[-1]
        self.parameters.Q.val =1
        self.set_distr_func(self._array_func)

    def _set_arrays(self,gamma_array,n_gamma_array):
        _ids = np.argsort(gamma_array)
        self._array_gamma = gamma_array[_ids]
        self._array_n_gamma = n_gamma_array[_ids]

    def _array_func(self,gamma):
        msk_nan=self._array_n_gamma>0
        f_interp = interpolate.interp1d(np.log10(self._array_gamma[msk_nan]), np.log10(self._array_n_gamma[msk_nan]), bounds_error=False, kind='linear')
        y = np.power(10., f_interp(np.log10(gamma)))
        msk_nan = np.isnan(y)
        y[msk_nan] = 0

        return y



class JetkernelEmittersDistribution(EmittersDistribution):

    def __init__(self,name,jet,gamma_grid_size=None,log_values=False,emitters_type='electrons'):

        super(JetkernelEmittersDistribution, self).__init__(name,
                                                            spectral_type=name,
                                                            skip_build=True,
                                                            emitters_type=emitters_type)
        self._build( jet, name, log_values, gamma_grid_size,emitters_type)

    def _build(self, jet, name, log_values, gamma_grid_size,emitters_type):
        self._set_N_distr_name(name)
        self._user_defined = False
        self._name = name
        self._log_values = log_values
        self._set_jet(jet, name,log_values=log_values, emitters_type=emitters_type)
        self.set_grid_size(gamma_grid_size)

        self._parameters_dict= self._build_emitters_distribution_dict(name, emitters_type=emitters_type)

    def _set_jet(self,jet, name, log_values=False, emitters_type='electrons'):
        self._jet = jet
        #jet.set_emitters_distribution(self, log_values=log_values, emitters_type=emitters_type)
        set_str_attr(jet._blob, 'DISTR', name)
        set_str_attr(jet._blob, 'PARTICLE', emitters_type)

    def _set_N_distr_name(self,name):
        self.available_N_distr = available_N_distr
        if name == 'from_array':
            pass
        elif name not in self.available_N_distr:
            raise RuntimeError ("electron distribution model %s not allowed" % name, "please choose among: ", self.available_N_distr)
        else:
            pass

    def set_grid_size(self,gamma_grid_size):
        if gamma_grid_size is not None:
            setattr(self._jet._blob,'gamma_grid_size' ,gamma_grid_size)

        self._set_blob()
        self._fill()

    def update(self):
        self._set_blob()
        self._fill()

    @staticmethod
    def available_distributions():
        for k in N_distr_descr.keys():
            print('%s: %s'%(k,N_distr_descr[k]))

    @classmethod
    def from_array(cls, jet, custom_Ne, emitters_type='electrons'):
        # TODO update for protons

        if emitters_type != 'electrons':
            raise RuntimeError('not implemented yet for', emitters_type, 'only for electrons')

        name = 'from_array'

        BlazarSED.build_Ne_custom(jet._blob, custom_Ne.size)
        Ne_custom_ptr = getattr(jet._blob, 'Ne_custom')
        gamma_custom_ptr = getattr(jet._blob, 'gamma_e_custom')
        jet._blob.N=1.0
        jet._blob.N_0 = 1.0
        jet.Norm_distr=0
        for ID in range(custom_Ne.size):
            BlazarSED.set_elec_custom_array(gamma_custom_ptr, jet._blob, custom_Ne.e_array[ID], ID)
            BlazarSED.set_elec_custom_array(Ne_custom_ptr, jet._blob, custom_Ne.n_array[ID], ID)

        setattr(jet._blob, 'gmin', custom_Ne.e_array[0])
        setattr(jet._blob, 'gmax', custom_Ne.e_array[-1])

        return cls(name, jet, custom_Ne.gamma_grid_size)

    def _set_blob(self):

        if self.emitters_type=='electrons':
            BlazarSED.InitNe(self._jet._blob)
            self._Ne_name, self._gammae_name = gamma_dic_e['electron_distr']
            self.Ne_ptr = getattr(self._jet._blob, self._Ne_name)
            self.e_gamma_ptr = getattr(self._jet._blob, self._gammae_name)

        elif self.emitters_type=='protons':
            BlazarSED.Init_Np_Ne_pp(self._jet._blob)
            self._Ne_name, self._gammae_name = gamma_dic_e['electron_distr']
            self._Np_name, self._gammap_name = gamma_dic_p['proton_distr']
            self._Q_inj_e_second_name, self._gammae_inj_sec_name = gamma_dic_pp_e_second['e_second_inj']

            self.Np_ptr = getattr(self._jet._blob, self._Np_name)
            self.p_gamma_ptr = getattr(self._jet._blob, self._gammap_name)
            self.Ne_ptr = getattr(self._jet._blob, self._Ne_name)
            self.e_gamma_ptr = getattr(self._jet._blob, self._gammae_name)

            self._Q_inj_e_second_ptr = getattr(self._jet._blob, self._Q_inj_e_second_name)
            self.e_inj_second_gamma_ptr = getattr(self._jet._blob, self._gammae_inj_sec_name)
        else:
            raise RuntimeError('')

    def _fill(self):
        size = self._jet._blob.gamma_grid_size
        self.gamma_e = np.zeros(size)
        self.n_gamma_e = np.zeros(size)
        self.gamma_p = np.zeros(size)
        self.n_gamma_p = np.zeros(size)
        self.gamma_e_second_inj = np.zeros(size)
        self.n_gamma_e_second_inj = np.zeros(size)

        for ID in range(size):
            self.gamma_e[ID] = BlazarSED.get_elec_array(self.e_gamma_ptr, self._jet._blob, ID)
            self.n_gamma_e[ID] = BlazarSED.get_elec_array(self.Ne_ptr, self._jet._blob, ID)

        if self.emitters_type=='protons':
            for ID in range(size):
                self.gamma_p[ID] = BlazarSED.get_elec_array(self.p_gamma_ptr, self._jet._blob, ID)
                self.n_gamma_p[ID] = BlazarSED.get_elec_array(self.Np_ptr, self._jet._blob, ID)

            for ID in range(size):
                self.gamma_e_second_inj[ID] = BlazarSED.get_elec_array(self.e_inj_second_gamma_ptr, self._jet._blob, ID)
                self.n_gamma_e_second_inj[ID] = BlazarSED.get_elec_array(self._Q_inj_e_second_ptr, self._jet._blob, ID)

        self.n_gamma_p[np.isnan(self.n_gamma_p)]=0
        self.n_gamma_p[np.isinf(self.n_gamma_p)]=0

        self.n_gamma_e[np.isnan(self.n_gamma_e)] = 0
        self.n_gamma_e[np.isinf(self.n_gamma_e)] = 0


    def _build_emitters_distribution_dict(self, distribution_name, emitters_type='electrons'):
        """
        Builds the dictionary to init the :class:`.JetParameter`
        objects   for the electron  distribution:

        The following :class:`.JetParameter`: objects the *do not* depend
        on the type of electron distribution

                - N, particle density in cm^-3

                - gmin, the minimum value of the electron Lorentz factor

                - gmax, the maximum value of the electron Lorentz factor

        The following :class:`.JetParameter`: objects *depend* on the type of electron distribution:

            - **power law**, electron_distribution='pl'

               - p

            - **broken power-law**, electron_distribution= **'bkn'**

                - p
                - p_1
                - gamma_break

            - **log-parabola**, electron_distribution= **'lp'**

                - r
                - s
                - gamma0_log_parab (fixed)

            - **log-parabola** with a low energy power-law tail, electron_distribution= **'lppl'**

                - r
                - s
                - gamma0_log_parab

            - **log-parabola** defined by peak energy, electron_distribution= **'lpep'**

                - r
                - s
                - gammap_log_parab,

            - **power-law cut-off**, lectron_distribution= **'plc'**

                - p
                - gamma_cut

        """

        model_dic, a_h, b_h, a_l, b_l, a_t, b_t= self._get_base_dict_and_bounds()

        a_h,b_h=self.set_bounds(1,1E15,log_val=self._log_values)
        a_l, b_l = self.set_bounds(1, 1E9, log_val=self._log_values)
        a_t, b_t = self.set_bounds(1, 1E9, log_val=self._log_values)

        if distribution_name == 'pl':
            model_dic['p'] = JetModelDictionaryPar(ptype='HE_spectral_slope',vmin=-10.,vmax=10.,punit='')

        if distribution_name == 'bkn':
            model_dic['p'] = JetModelDictionaryPar(ptype='LE_spectral_slope',vmin=-10.,vmax=10.,punit='')
            model_dic['p_1'] = JetModelDictionaryPar(ptype='HE_spectral_slope',vmin=-10.,vmax=10.,punit='')
            model_dic['gamma_break'] = JetModelDictionaryPar(ptype='turn-over-energy',vmin=a_t,vmax=b_t,punit='lorentz-factor',log=self._log_values)

        if distribution_name == 'lp':
            model_dic['s'] = JetModelDictionaryPar(ptype='LE_spectral_slope', vmin=-10., vmax=10., punit='')
            model_dic['r'] = JetModelDictionaryPar(ptype='spectral_curvature', vmin=-15., vmax=15., punit='')
            model_dic['gamma0_log_parab'] = JetModelDictionaryPar(ptype='turn-over-energy', vmin=a_t, vmax=b_t,punit='lorentz-factor', log=self._log_values,froz=True)

        if distribution_name == 'lppl':
            model_dic['s'] = JetModelDictionaryPar(ptype='LE_spectral_slope',vmin=-10.,vmax=10.,punit='')
            model_dic['r'] = JetModelDictionaryPar(ptype='spectral_curvature',vmin=-15.,vmax=15.,punit='')
            model_dic['gamma0_log_parab'] = JetModelDictionaryPar(ptype='turn-over-energy',vmin=a_t,vmax=b_t,punit='lorentz-factor',log=self._log_values)

        if distribution_name == 'lpep':
            model_dic['r'] =  JetModelDictionaryPar(ptype='spectral_curvature',vmin=-15.,vmax=15.,punit='')
            model_dic['gammap_log_parab'] = JetModelDictionaryPar(ptype='turn-over-energy',vmin=a_t,vmax=b_t,punit='lorentz-factor',log=self._log_values)

        if distribution_name == 'plc':
            model_dic['p'] = JetModelDictionaryPar(ptype='LE_spectral_slope', vmin=-10., vmax=10., punit='')
            model_dic['gamma_cut'] = JetModelDictionaryPar(ptype='turn-over-energy',vmin=a_t,vmax=b_t,punit='lorentz-factor',log=self._log_values)

        if distribution_name == 'spitkov':
            model_dic['spit_index'] = JetModelDictionaryPar(ptype='LE_spectral_slope',vmin=-10.,vmax=10.,punit='')
            model_dic['spit_temp'] = JetModelDictionaryPar(ptype='turn-over-energy',vmin=a_t,vmax=b_t,punit='lorentz-factor',log=self._log_values)
            model_dic['spit_gamma_th'] =JetModelDictionaryPar(ptype='turn-over-energy',vmin=a_t,vmax=b_t,punit='lorentz-factor',log=self._log_values)

        if distribution_name == 'lppl_pile_up':
            model_dic['s'] =JetModelDictionaryPar(ptype='LE_spectral_slope',vmin=-10.,vmax=10.,punit='')
            model_dic['r'] =JetModelDictionaryPar(ptype='spectral_curvature',vmin=-15.,vmax=15.,punit='')
            model_dic['gamma0_log_parab'] = JetModelDictionaryPar(ptype='turn-over-energy',vmin=a_t,vmax=b_t,punit='lorentz-factor',log=self._log_values)
            model_dic['gamma_inj'] = JetModelDictionaryPar(ptype='turn-over-energy',vmin=a_t,vmax=b_t,punit='lorentz-factor',log=self._log_values)
            model_dic['gamma_pile_up'] = JetModelDictionaryPar(ptype='turn-over-energy',vmin=a_t,vmax=b_t,punit='lorentz-factor',log=self._log_values)
            model_dic['ratio_pile_up'] = JetModelDictionaryPar(ptype='turn-over-energy',vmin=0.0,vmax=None,punit='')
            model_dic['alpha_pile_up'] = JetModelDictionaryPar(ptype='turn-over-energy',vmin=0.0,vmax=10.0,punit='')

        if distribution_name == 'bkn_pile_up':
            model_dic['p'] = JetModelDictionaryPar(ptype='LE_spectral_slope',vmin=-10.,vmax=10.,punit='')
            model_dic['p_1'] = JetModelDictionaryPar(ptype='LE_spectral_slope',vmin=-10.,vmax=10.,punit='')
            model_dic['gamma_break'] =  JetModelDictionaryPar(ptype='turn-over-energy',vmin=a_t,vmax=b_t,punit='lorentz-factor',log=self._log_values)
            model_dic['gamma_pile_up'] = JetModelDictionaryPar(ptype='turn-over-energy',vmin=a_t,vmax=b_t,punit='lorentz-factor',log=self._log_values)
            model_dic['gamma_pile_up_cut'] =JetModelDictionaryPar(ptype='turn-over-energy',vmin=a_t,vmax=b_t,punit='lorentz-factor',log=self._log_values)
            model_dic['alpha_pile_up'] = JetModelDictionaryPar(ptype='turn-over-energy',vmin=0.0,vmax=10.0,punit='')

        if distribution_name == 'from_array':
            model_dic.pop('gmin')
            model_dic.pop('gmax')

        return model_dic







__author__ = "Andrea Tramacere"



from scipy import interpolate

import numpy as np

from astropy.units import Unit as u
from astropy.units import spectral
from astropy.table import Table

import os

from  .plot_sedfit import PlotSED,PlotSpectralMultipl

from .model_parameters import ModelParameter
from .base_model import  Model, MultiplicativeModel

__all__=['TemplateTable2D', 'EBLAbsorptionTemplate']



class TemplateTable2D(Model):
    """
    Class to handle spectral templates
    """
    def __init__(self,
                 x_values,
                 y_values,
                 z_values,
                 log_input_grid=False,
                 log_log_interp=False,
                 x_in_units=None,
                 y_in_units=None,
                 x_out_units='Hz',
                 y_out_units='erg cm-2 s-1',
                 z_in_units=None,
                 z_out_units=None,
                 nu_size=100,
                 name='TableModel2D',
                 zero=1E-100):
        """
        """
        super(TemplateTable2D, self).__init__(name=name)


        self._zero=zero

        self._log_log_interp = log_log_interp

        if log_input_grid is False:

            self.x_values = x_values
            self.y_values = y_values
            self.z_values = z_values
        else:
            self.z_values = np.power(10., z_values)
            self.x_values = np.power(10., x_values)
            self.y_values = np.power(10., y_values)

        if x_in_units is not None:
            self.x_values *= u(x_in_units).to(x_out_units, equivalencies=spectral())

        if hasattr(self.x_values,'value'):
            _x = self.x_values.value
        else:
            _x = self.x_values

        if y_in_units is not None:
            self.y_values *= u(y_in_units).to(y_out_units, equivalencies=spectral())

        if hasattr(self.y_values,'value'):
            _y = self.y_values.value
        else:
            _y = self.y_values

        if z_in_units is not None:
            self.z_values *= u(z_in_units).to(z_out_units, equivalencies=spectral())

        if hasattr(self.z_values, 'value'):
            _z = self.z_values.value
        else:
            _z = self.z_values

        if log_log_interp is True:
            self._scale='log-log'
            _x[_x<self._zero]=self._zero
            self._x_grid = np.log10(_x)
            self._y_grid = np.log10(_y)
            self._z_grid = np.log10(_z)

        else:
            self._scale='lin-lin'
            self._x_grid =  _x
            self._y_grid =  _y
            self._z_grid =  _z


        self.interp_func = interpolate.RectBivariateSpline(self._x_grid, self._y_grid, self._z_grid)

        self.nu_size = nu_size
        
        self.name = name
        
        self.model_type = 'table2D'

    def plot_model(self,plot_obj=None,clean=False,label=None,sed_data=None,color=None, density=False,frame='obs'):
        plot_obj=self._set_up_plot(plot_obj,sed_data,frame,density)

        if clean==True:
            plot_obj.clean_model_lines()

        if label is None:
            label=self.name

        plot_obj.add_model_plot(self.SED, line_style='-', label=label, flim=self.flux_plot_lim,color=color, density=density,frame=frame)

        return plot_obj

    def _func(self,x,y):

        if np.shape(x)==():
                x=np.array([x])

        if self._log_log_interp is True:
            x=np.log10(x)
            y=np.log10(y)
             
        model = self.interp_func(x,y)
        #
        model[model<self._zero]=self._zero

        if self._log_log_interp is True:
            return np.power(10., model)
        else:
            pass

        return model

    
    def eval(self,fill_SED=True,x=None,y=None,get_model=False,loglog=False):
        """
        Evaluates the Template for the current parameters values
        """    

        if x is None:
            x = np.copy(self.x_values)

        if loglog is True:
            x = np.power(10,x)

        if y is None:
            y = np.copy(self.y_values)

        if loglog is False:
            y = np.power(10,y)
        
        

        
        model=self._func(x,y)



        if get_model==True:
            if loglog==False:
            
                return model
            else:
                
                return np.log10(model)

        else:
            return None


class EBLAbsorptionTemplate(TemplateTable2D,MultiplicativeModel):
    """

    """
    def __init__(self,
                 redshift_array,
                 energy_array,
                 tau_values,
                 applied_model=None,
                 z=1.0,
                 template_name='tau_ebl_template',
                 nu_size=100):
        """

        Parameters
        ----------
        redshift_array
        energy_array
        tau_values
        applied_model
        z
        template_name
        nu_size
        """
        super(EBLAbsorptionTemplate, self).__init__(
            x_values=redshift_array,
            y_values=energy_array,
            z_values=tau_values,
            log_input_grid=False,
            log_log_interp=False,
            nu_size=nu_size,
            name=template_name
        )



        if z is None and applied_model is not None:
            self.parameters.add_par(applied_model.get_par_by_type('redshift'))
        elif applied_model is None and z is not None:
            self.parameters.add_par(ModelParameter(name='z_cosm',
                                                   par_type='redshift',
                                                   val=z,
                                                   val_min=0,
                                                   val_max=None,
                                                   units='',
                                                   frozen=True,
                                                   log=False))
        elif  z is not None and applied_model is not None:
            raise RuntimeError('either you provide a redshift value (z) or model with redshift parameter')

    @property
    def z(self):
        if hasattr(self,'_z'):
            if isinstance(self._z,ModelParameter):
                return self._z.val
            else:
                return self._z
        else:
            return  None

    @z.setter
    def z(self,z):
        self._z = z

    def apply_to(self,model):
        p_m = model.parameters.get_par_by_type('redshift')
        _p= self.get_par_by_name('redshift')
        _p.val=p_m.val

    def _check(self):
        pass

    @classmethod
    def from_name(cls,template_name,applied_model=None,z=1.0,nu_size=100):
        """

        Parameters
        ----------
        template_name
        applied_model
        z
        nu_size

        Returns
        -------

        """

        _Templates_dir = os.path.dirname(__file__) + '/ebl_data'

        _allowed_templates = ['Finke_2010', 'Dominguez_2010', 'Franceschini_2008']

        if template_name not in _allowed_templates:
            raise ValueError('template EBL model', template_name, 'not in allowdr',_allowed_templates)

        _template_name_dict = {}

        _template_name_dict['Finke_2010'] = 'tau_finke_2010.fits'
        _template_name_dict['Dominguez_2010'] = 'tau_dominguez_2010.fits'
        _template_name_dict['Franceschini_2008'] = 'tau_franceschini_2008.dat'

        file_path = os.path.join(_Templates_dir, _template_name_dict[template_name])

        if file_path.endswith('fits'):
            data = Table.read(file_path,format='fits')
        elif file_path.endswith('dat'):
            data = Table.read(file_path, format='ascii.ecsv')
        else:
            data = Table.read(file_path)

        y_values = data['energies']
        y_values = np.log10(y_values.to('Hz', equivalencies=spectral()).value)

        try:
            x_values = np.array(data.meta['REDSHIFT'], dtype=np.float)
        except:
            x_values = np.array(data.meta['redshift'], dtype=np.float)

        cn = [name for name in data.colnames if name != 'energies']
        z_values = np.array([data[n].data for n in cn])

        return cls(redshift_array=x_values,
                   energy_array=y_values,
                   tau_values=z_values,
                   template_name=template_name,
                   applied_model=applied_model,
                   z=z,
                   nu_size=nu_size)

    def eval(self, fill_SED=True, nu=None, get_model=False, loglog=False):
        """

        Parameters
        ----------
        fill_SED
        nu
        get_model
        loglog
        z

        Returns
        -------

        """

        out_model=None

        if nu is None:
            log_nu = np.copy(self.y_values)

        else:
            if loglog is True:
                log_nu = nu
            else:
                log_nu = np.log10(nu)


        z=self.parameters.get_par_by_name('z_cosm').val

        model = np.exp(-self._func(z,log_nu).T).flatten()

        if get_model == True:
            if loglog == False:
                out_model=model
            else:
                out_model=  np.log10(model)

        if fill_SED is True:
            if loglog is False:
                _nu=np.power(10.,log_nu)
            else:
                _nu=np.copy(log_nu)

            self.nu=_nu
            self.tau=model

        return out_model

    def plot_model(self, plot_obj=None,  label=None, line_style='-',color=None,frame='obs'):
        if plot_obj is None:
            plot_obj = PlotSpectralMultipl()

        if label is None:
            label = '%s'%self.name
            label += ' ,z=%5.5f'%self.parameters.get_par_by_name('z_cosm').val

        plot_obj.plot(nu=self.nu,y=self.tau, y_label=r'$ log10(\exp^{- \tau}) $', line_style=line_style, label=label, color=color)


        return plot_obj
